import numpy as np

from ..base import EphemeralNode
from ..base import FunctionalNode
from ..base import TerminalNode
from ..base import UniversalSet
from ..optimizers import DifferentialEvolution
from ..optimizers import GeneticAlgorithm
from ..optimizers import GeneticProgramming
from ..optimizers import SHADE
from ..optimizers import SHAGA
from ..optimizers import SelfCGA
from ..optimizers import SelfCGP
from ..optimizers import jDE
from ..tools.metrics import coefficient_determination
from ..tools.operators import Add
from ..tools.operators import Div
from ..tools.operators import Mul
from ..tools.operators import Neg
from ..tools.random import binary_string_population
from ..tools.random import float_population
from ..tools.random import half_and_half


def test_GeneticAlgorithm_start_settings():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 100
    pop_size = 50
    str_len = 200

    # simple start
    optimizer = GeneticAlgorithm(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=True,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["fitness_max"]) == iters
    for i in range(len(stats["fitness_max"][:-1])):
        assert stats["fitness_max"][i] <= stats["fitness_max"][i + 1]
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15
    optimizer = GeneticAlgorithm(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = GeneticAlgorithm(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)


def test_GeneticAlgorithm_set_strategy():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 3
    pop_size = 8
    str_len = 5

    initial_population = binary_string_population(pop_size=pop_size, str_len=str_len)
    selections = (
        "proportional",
        "rank",
        "tournament_k",
        "tournament_3",
        "tournament_5",
        "tournament_7",
    )
    crossover = (
        "empty",
        "one_point",
        "two_point",
        "uniform_2",
        "uniform_7",
        "uniform_k",
        "uniform_prop_2",
        "uniform_prop_7",
        "uniform_prop_k",
        "uniform_rank_2",
        "uniform_rank_7",
        "uniform_rank_k",
        "uniform_tour_3",
        "uniform_tour_7",
        "uniform_tour_k",
    )

    mutation = ("weak", "average", "strong", "custom_rate")

    tour_size = (3, 4, 5)
    parents_num = (7, 8, 9)
    mutation_rate = (0.11, 0.22)

    for selections_i in selections:
        for crossover_i in crossover:
            for mutation_i in mutation:
                for tour_size_i in tour_size:
                    for parents_num_i in parents_num:
                        for mutation_rate_i in mutation_rate:
                            optimizer = GeneticAlgorithm(
                                fitness_function,
                                iters=iters,
                                pop_size=pop_size,
                                str_len=str_len,
                                optimal_value=None,
                                termination_error_value=0,
                                no_increase_num=None,
                                minimization=False,
                                show_progress_each=None,
                                keep_history=True,
                                init_population=initial_population,
                                selection=selections_i,
                                crossover=crossover_i,
                                mutation=mutation_i,
                                tour_size=tour_size_i,
                                elitism=True,
                                parents_num=parents_num_i,
                                mutation_rate=mutation_rate_i,
                            )

                            assert optimizer._specified_selection == selections_i
                            assert optimizer._specified_crossover == crossover_i
                            assert optimizer._specified_mutation == mutation_i
                            assert optimizer._tour_size == tour_size_i
                            assert optimizer._parents_num == parents_num_i
                            assert optimizer._mutation_rate == mutation_rate_i

                            if selections_i[-2:] == "_k":
                                assert (
                                    optimizer._selection_pool[optimizer._specified_selection][1]
                                    == tour_size_i
                                )
                            if crossover_i[-2:] == "_k":
                                assert (
                                    optimizer._crossover_pool[optimizer._specified_crossover][1]
                                    == parents_num_i
                                )
                            if mutation_i == "custom_rate":
                                assert (
                                    optimizer._mutation_pool[optimizer._specified_mutation][1]
                                    == mutation_rate_i
                                )

                            optimizer.fit()

                            stats = optimizer.get_stats()
                            assert np.all(stats["population_g"][0] == initial_population)


def test_SelfCGA_start_settings():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 100
    pop_size = 50
    str_len = 200

    # simple start
    optimizer = SelfCGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=True,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["fitness_max"]) == iters
    for i in range(len(stats["fitness_max"][:-1])):
        assert stats["fitness_max"][i] <= stats["fitness_max"][i + 1]
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15
    optimizer = SelfCGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = SelfCGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)


def test_SelfCGA_set_strategy():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 10
    pop_size = 10
    str_len = 5
    initial_population = binary_string_population(pop_size=pop_size, str_len=str_len)
    tour_size = 8
    parents_num = 8
    mutation_rate = 0.5

    optimizer = SelfCGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=None,
        keep_history=True,
        init_population=initial_population,
        selections=(
            "proportional",
            "rank",
            "tournament_k",
            "tournament_3",
        ),
        crossovers=(
            "empty",
            "one_point",
            "two_point",
            "uniform_2",
            "uniform_k",
            "uniform_prop_2",
            "uniform_prop_k",
            "uniform_rank_2",
            "uniform_rank_k",
            "uniform_tour_3",
            "uniform_tour_k",
        ),
        mutations=("weak", "custom_rate"),
        tour_size=tour_size,
        elitism=True,
        parents_num=parents_num,
        mutation_rate=mutation_rate,
    )
    optimizer.fit()
    stats = optimizer.get_stats()
    assert np.all(stats["population_g"][0] == initial_population)


def test_GeneticProgramming_start_settings():
    def generator1():
        return np.round(np.random.uniform(0, 10), 4)

    def generator2():
        return np.random.randint(0, 10)

    def problem(x):
        return 3 * x[:, 0] ** 2 + 2 * x[:, 0] + 5

    function = problem
    left_border = -4.5
    right_border = 4.5
    sample_size = 300
    n_dimension = 1

    iters = 20
    pop_size = 15

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)]
    ).T
    y = function(X)

    functional_set = (
        FunctionalNode(Add()),
        FunctionalNode(Mul()),
        FunctionalNode(Neg()),
        FunctionalNode(Div()),
    )

    terminal_set = [TerminalNode(X[:, i], f"x{i}") for i in range(n_dimension)]
    terminal_set.extend([EphemeralNode(generator1), EphemeralNode(generator2)])
    uniset = UniversalSet(functional_set, terminal_set)

    def fitness_function(trees):
        fitness = []
        for tree in trees:
            y_pred = tree() * np.ones(len(y))
            fitness.append(-coefficient_determination(y, y_pred))
        return np.array(fitness)

    # simple start
    optimizer = GeneticProgramming(
        fitness_function=fitness_function,
        uniset=uniset,
        pop_size=pop_size,
        iters=iters,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        show_progress_each=1,
        minimization=True,
        keep_history=True,
    )

    optimizer.fit()

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["fitness_max"]) == iters
    for i in range(len(stats["fitness_max"][:-1])):
        assert stats["fitness_max"][i] <= stats["fitness_max"][i + 1]
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15

    optimizer = GeneticProgramming(
        fitness_function=fitness_function,
        uniset=uniset,
        pop_size=pop_size,
        iters=iters,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        show_progress_each=1,
        minimization=False,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = GeneticProgramming(
        fitness_function=fitness_function,
        uniset=uniset,
        pop_size=pop_size,
        iters=iters,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        show_progress_each=1,
        minimization=False,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)


def test_GeneticProgramming_set_strategy():
    def generator1():
        return np.round(np.random.uniform(0, 10), 4)

    def generator2():
        return np.random.randint(0, 10)

    def problem(x):
        return 3 * x[:, 0] ** 2 + 2 * x[:, 0] + 5

    function = problem
    left_border = -4.5
    right_border = 4.5
    sample_size = 30
    n_dimension = 1

    iters = 2
    pop_size = 10

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)]
    ).T
    y = function(X)

    functional_set = (
        FunctionalNode(Add()),
        FunctionalNode(Mul()),
        FunctionalNode(Neg()),
        FunctionalNode(Div()),
    )

    terminal_set = [TerminalNode(X[:, i], f"x{i}") for i in range(n_dimension)]
    terminal_set.extend([EphemeralNode(generator1), EphemeralNode(generator2)])
    uniset = UniversalSet(functional_set, terminal_set)

    def fitness_function(trees):
        fitness = []
        for tree in trees:
            y_pred = tree() * np.ones(len(y))
            fitness.append(-coefficient_determination(y, y_pred))
        return np.array(fitness)

    initial_population = half_and_half(pop_size=pop_size, uniset=uniset, max_level=14)
    selections = (
        "proportional",
        "rank",
        "tournament_k",
        "tournament_3",
    )
    crossover = (
        "empty",
        "gp_standart",
        "gp_one_point",
        "gp_uniform_2",
        "gp_uniform_k",
        "gp_uniform_prop_2",
        "gp_uniform_prop_k",
        "gp_uniform_rank_2",
        "gp_uniform_rank_k",
        "gp_uniform_tour_3",
        "gp_uniform_tour_k",
    )

    mutation = (
        "gp_weak_point",
        "gp_custom_rate_point",
        "gp_weak_grow",
        "gp_custom_rate_grow",
        "gp_weak_swap",
        "gp_custom_rate_swap",
        "gp_weak_shrink",
        "gp_custom_rate_shrink",
    )

    for selections_i in selections:
        for crossover_i in crossover:
            for mutation_i in mutation:
                optimizer = GeneticProgramming(
                    fitness_function,
                    iters=iters,
                    pop_size=pop_size,
                    uniset=uniset,
                    optimal_value=None,
                    termination_error_value=0,
                    no_increase_num=None,
                    minimization=False,
                    show_progress_each=None,
                    keep_history=True,
                    init_population=initial_population,
                    selection=selections_i,
                    crossover=crossover_i,
                    mutation=mutation_i,
                    tour_size=3,
                    elitism=True,
                    parents_num=3,
                    mutation_rate=0.33,
                )

                assert optimizer._specified_selection == selections_i
                assert optimizer._specified_crossover == crossover_i
                assert optimizer._specified_mutation == mutation_i
                assert optimizer._tour_size == 3
                assert optimizer._parents_num == 3
                assert optimizer._mutation_rate == 0.33

                if selections_i[-2:] == "_k":
                    assert optimizer._selection_pool[optimizer._specified_selection][1] == 3
                if crossover_i[-2:] == "_k":
                    assert optimizer._crossover_pool[optimizer._specified_crossover][1] == 3
                if mutation_i[:9] == "gp_custom":
                    assert optimizer._mutation_pool[optimizer._specified_mutation][1] == 0.33

                optimizer.fit()

                stats = optimizer.get_stats()
                assert np.all(stats["population_g"][0] == initial_population)


def test_SelfCGP_start_settings():
    def generator1():
        return np.round(np.random.uniform(0, 10), 4)

    def generator2():
        return np.random.randint(0, 10)

    def problem(x):
        return 3 * x[:, 0] ** 2 + 2 * x[:, 0] + 5

    function = problem
    left_border = -4.5
    right_border = 4.5
    sample_size = 300
    n_dimension = 1

    iters = 20
    pop_size = 15

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)]
    ).T
    y = function(X)

    functional_set = (
        FunctionalNode(Add()),
        FunctionalNode(Mul()),
        FunctionalNode(Neg()),
        FunctionalNode(Div()),
    )

    terminal_set = [TerminalNode(X[:, i], f"x{i}") for i in range(n_dimension)]
    terminal_set.extend([EphemeralNode(generator1), EphemeralNode(generator2)])
    uniset = UniversalSet(functional_set, terminal_set)

    def fitness_function(trees):
        fitness = []
        for tree in trees:
            y_pred = tree() * np.ones(len(y))
            fitness.append(-coefficient_determination(y, y_pred))
        return np.array(fitness)

    # simple start
    optimizer = SelfCGP(
        fitness_function=fitness_function,
        uniset=uniset,
        pop_size=pop_size,
        iters=iters,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        show_progress_each=1,
        minimization=True,
        keep_history=True,
    )

    optimizer.fit()

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["fitness_max"]) == iters
    for i in range(len(stats["fitness_max"][:-1])):
        assert stats["fitness_max"][i] <= stats["fitness_max"][i + 1]
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15

    optimizer = SelfCGP(
        fitness_function=fitness_function,
        uniset=uniset,
        pop_size=pop_size,
        iters=iters,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        show_progress_each=1,
        minimization=False,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = SelfCGP(
        fitness_function=fitness_function,
        uniset=uniset,
        pop_size=pop_size,
        iters=iters,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        show_progress_each=1,
        minimization=False,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)


def test_SelfCGP_set_strategy():
    def generator1():
        return np.round(np.random.uniform(0, 10), 4)

    def generator2():
        return np.random.randint(0, 10)

    def problem(x):
        return 3 * x[:, 0] ** 2 + 2 * x[:, 0] + 5

    function = problem
    left_border = -4.5
    right_border = 4.5
    sample_size = 300
    n_dimension = 1

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)]
    ).T
    y = function(X)

    functional_set = (
        FunctionalNode(Add()),
        FunctionalNode(Mul()),
        FunctionalNode(Neg()),
        FunctionalNode(Div()),
    )

    terminal_set = [TerminalNode(X[:, i], f"x{i}") for i in range(n_dimension)]
    terminal_set.extend([EphemeralNode(generator1), EphemeralNode(generator2)])
    uniset = UniversalSet(functional_set, terminal_set)

    def fitness_function(trees):
        fitness = []
        for tree in trees:
            y_pred = tree() * np.ones(len(y))
            fitness.append(-coefficient_determination(y, y_pred))
        return np.array(fitness)

    iters = 10
    pop_size = 10
    initial_population = half_and_half(pop_size=pop_size, uniset=uniset, max_level=14)
    tour_size = 8
    parents_num = 8
    mutation_rate = 0.5

    optimizer = SelfCGP(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        uniset=uniset,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=None,
        keep_history=True,
        init_population=initial_population,
        selections=(
            "proportional",
            "rank",
            "tournament_k",
            "tournament_3",
            "tournament_5",
            "tournament_7",
        ),
        crossovers=(
            "empty",
            "gp_standart",
            "gp_one_point",
            "gp_uniform_2",
            "gp_uniform_7",
            "gp_uniform_k",
            "gp_uniform_prop_2",
            "gp_uniform_prop_7",
            "gp_uniform_prop_k",
            "gp_uniform_rank_2",
            "gp_uniform_rank_7",
            "gp_uniform_rank_k",
            "gp_uniform_tour_3",
            "gp_uniform_tour_7",
            "gp_uniform_tour_k",
        ),
        mutations=(
            "gp_weak_point",
            "gp_average_point",
            "gp_strong_point",
            "gp_custom_rate_point",
            "gp_weak_grow",
            "gp_average_grow",
            "gp_strong_grow",
            "gp_custom_rate_grow",
            "gp_weak_swap",
            "gp_average_swap",
            "gp_strong_swap",
            "gp_custom_rate_swap",
            "gp_weak_shrink",
            "gp_average_shrink",
            "gp_strong_shrink",
            "gp_custom_rate_shrink",
        ),
        tour_size=tour_size,
        elitism=True,
        parents_num=parents_num,
        mutation_rate=mutation_rate,
    )
    optimizer.fit()
    stats = optimizer.get_stats()
    assert np.all(stats["population_g"][0] == initial_population)


def test_DifferentialEvolution_start_settings():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 100
    pop_size = 50
    n_vars = 10
    left = np.full(n_vars, -1, dtype=np.float64)
    right = np.full(n_vars, 1, dtype=np.float64)

    # simple start
    optimizer = DifferentialEvolution(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left=left,
        right=right,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=True,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["fitness_max"]) == iters
    for i in range(len(stats["fitness_max"][:-1])):
        assert stats["fitness_max"][i] <= stats["fitness_max"][i + 1]
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15
    optimizer = DifferentialEvolution(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left=left,
        right=right,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = DifferentialEvolution(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left=left,
        right=right,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)


def test_DifferentialEvolution_set_strategy():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 2
    pop_size = 10
    n_vars = 10
    left = np.full(n_vars, -1, dtype=np.float64)
    right = np.full(n_vars, 1, dtype=np.float64)
    initial_population = float_population(pop_size=pop_size, left=left, right=right)

    mutations = ("best_1", "rand_1", "current_to_best_1", "rand_to_best1", "best_2", "rand_2")
    F = (0.5, 0.3)
    CR = (0.31, 0.33)

    for mutation_i in mutations:
        for F_i in F:
            for CR_i in CR:
                optimizer = DifferentialEvolution(
                    fitness_function,
                    iters=iters,
                    pop_size=pop_size,
                    left=left,
                    right=right,
                    optimal_value=None,
                    termination_error_value=0,
                    no_increase_num=None,
                    minimization=True,
                    show_progress_each=1,
                    keep_history=True,
                    mutation=mutation_i,
                    init_population=initial_population,
                    F=F_i,
                    CR=CR_i,
                )
                optimizer.fit()
                stats = optimizer.get_stats()

                assert optimizer._specified_mutation == mutation_i
                assert optimizer._F == F_i
                assert optimizer._CR == CR_i
                print(stats.keys())
                assert np.all(stats["population_g"][0] == initial_population)


def test_jDE_start_settings():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 100
    pop_size = 50
    n_vars = 10
    left = np.full(n_vars, -1, dtype=np.float64)
    right = np.full(n_vars, 1, dtype=np.float64)

    # simple start
    optimizer = jDE(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left=left,
        right=right,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=True,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["fitness_max"]) == iters
    for i in range(len(stats["fitness_max"][:-1])):
        assert stats["fitness_max"][i] <= stats["fitness_max"][i + 1]
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15
    optimizer = jDE(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left=left,
        right=right,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = jDE(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left=left,
        right=right,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)


def test_jDE_set_strategy():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 2
    pop_size = 10
    n_vars = 10
    left = np.full(n_vars, -1, dtype=np.float64)
    right = np.full(n_vars, 1, dtype=np.float64)
    initial_population = float_population(pop_size=pop_size, left=left, right=right)

    mutations = ("best_1", "rand_1", "current_to_best_1", "rand_to_best1", "best_2", "rand_2")
    F_min = (0.05, 0.1)
    F_max = (0.9, 0.95)
    t_F = (0.051, 0.21)
    t_CR = (0.05, 0.1)

    for mutation_i in mutations:
        for F_min_i in F_min:
            for F_max_i in F_max:
                for t_F_i in t_F:
                    for t_CR_i in t_CR:
                        optimizer = jDE(
                            fitness_function,
                            iters=iters,
                            pop_size=pop_size,
                            left=left,
                            right=right,
                            optimal_value=None,
                            termination_error_value=0,
                            no_increase_num=None,
                            minimization=True,
                            show_progress_each=1,
                            keep_history=True,
                            mutation=mutation_i,
                            F_min=F_min_i,
                            F_max=F_max_i,
                            t_CR=t_CR_i,
                            t_F=t_F_i,
                            init_population=initial_population,
                        )
                        optimizer.fit()
                        stats = optimizer.get_stats()

                        assert optimizer._specified_mutation == mutation_i
                        assert optimizer._F_min == F_min_i
                        assert optimizer._F_max == F_max_i
                        assert optimizer._t_CR == t_CR_i
                        assert optimizer._t_F == t_F_i
                        print(stats.keys())
                        assert np.all(stats["population_g"][0] == initial_population)


def test_SHADE_start_settings():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 100
    pop_size = 50
    n_vars = 10
    left = np.full(n_vars, -1, dtype=np.float64)
    right = np.full(n_vars, 1, dtype=np.float64)

    # simple start
    optimizer = SHADE(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left=left,
        right=right,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=True,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["fitness_max"]) == iters
    for i in range(len(stats["fitness_max"][:-1])):
        assert stats["fitness_max"][i] <= stats["fitness_max"][i + 1]
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15
    optimizer = SHADE(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left=left,
        right=right,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = SHADE(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        left=left,
        right=right,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)


def test_SHAGA_start_settings():
    def fitness_function(x):
        return np.sum(x, axis=1, dtype=np.float64)

    iters = 100
    pop_size = 50
    str_len = 200

    # simple start
    optimizer = SHAGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=None,
        minimization=True,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    fittest = optimizer.get_fittest()
    assert isinstance(fittest, dict)

    stats = optimizer.get_stats()

    assert optimizer.get_remains_calls() == 0
    assert len(stats["fitness_max"]) == iters
    for i in range(len(stats["fitness_max"][:-1])):
        assert stats["fitness_max"][i] <= stats["fitness_max"][i + 1]
    assert optimizer._sign == -1

    # start with the no_increase_num is equal 15
    def fitness_function(x):
        return np.ones(len(x), dtype=np.float64)

    no_increase_num = 15
    optimizer = SHAGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=None,
        termination_error_value=0,
        no_increase_num=no_increase_num,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()

    assert optimizer.get_remains_calls() == pop_size * (iters - no_increase_num - 1)
    assert optimizer._sign == 1

    # start with the optimal_value is equal 1
    optimizer = SHAGA(
        fitness_function,
        iters=iters,
        pop_size=pop_size,
        str_len=str_len,
        optimal_value=1,
        termination_error_value=0,
        no_increase_num=None,
        minimization=False,
        show_progress_each=1,
        keep_history=True,
    )

    optimizer.fit()
    assert optimizer.get_remains_calls() == pop_size * (iters - 1)



import numpy as np

from ..benchmarks import BanknoteDataset
from ..benchmarks import BreastCancerDataset
from ..benchmarks import CreditRiskDataset
from ..benchmarks import DigitsDataset
from ..benchmarks import IrisDataset
from ..benchmarks import OneMax
from ..benchmarks import UserKnowladgeDataset
from ..benchmarks import WineDataset
from ..benchmarks.symbolicregression17 import (
    F1,
    F2,
    F3,
    F4,
    F5,
    F6,
    F7,
    F8,
    F9,
    F10,
    F11,
    F12,
    F13,
    F14,
    F15,
    F16,
    F17,
)

from ..benchmarks.CEC2005 import problems_dict
from ..optimizers import DifferentialEvolution
from ..optimizers import GeneticAlgorithm
from ..tools.random import float_population


def test_IrisDataset():
    data = IrisDataset()

    X = data.get_X()
    y = data.get_y()

    assert X.shape == (150, 4)
    assert y.shape == (150,)

    X_names = data.get_X_names()
    y_names = data.get_y_names()

    assert len(X_names) == 4
    assert len(y_names) == 3


def test_WineDataset():
    data = WineDataset()

    X = data.get_X()
    y = data.get_y()

    assert X.shape == (178, 13)
    assert y.shape == (178,)

    X_names = data.get_X_names()
    y_names = data.get_y_names()

    assert len(X_names) == 13
    assert len(y_names) == 3


def test_BreastCancerDataset():
    data = BreastCancerDataset()

    X = data.get_X()
    y = data.get_y()

    assert X.shape == (569, 30)
    assert y.shape == (569,)

    X_names = data.get_X_names()
    y_names = data.get_y_names()

    assert len(X_names) == 30
    assert len(y_names) == 2


def test_DigitsDataset():
    data = DigitsDataset()

    X = data.get_X()
    y = data.get_y()

    assert X.shape == (5620, 64)
    assert y.shape == (5620,)

    X_names = data.get_X_names()
    y_names = data.get_y_names()

    assert len(X_names) == 64
    assert len(y_names) == 10


def test_CreditRiskDataset():
    data = CreditRiskDataset()

    X = data.get_X()
    y = data.get_y()

    assert X.shape == (2000, 3)
    assert y.shape == (2000,)

    X_names = data.get_X_names()
    y_names = data.get_y_names()

    assert len(X_names) == 3
    assert len(y_names) == 2


def test_UserKnowladgeDataset():
    data = UserKnowladgeDataset()

    X = data.get_X()
    y = data.get_y()

    assert X.shape == (403, 5)
    assert y.shape == (403,)

    X_names = data.get_X_names()
    y_names = data.get_y_names()

    assert len(X_names) == 5
    assert len(y_names) == 4


def test_BanknoteDataset():
    data = BanknoteDataset()

    X = data.get_X()
    y = data.get_y()

    assert X.shape == (1372, 4)
    assert y.shape == (1372,)

    X_names = data.get_X_names()
    y_names = data.get_y_names()

    assert len(X_names) == 4
    assert len(y_names) == 2


def test_symbolicregression17():
    left_border = -4.5
    right_border = 4.5
    sample_size = 30
    n_dimension = 1

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)],
        dtype=np.float64,
    ).T

    F1(X)
    F2(X)
    F16(X)
    F17(X)

    n_dimension = 2

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)],
        dtype=np.float64,
    ).T

    F3(X)
    F4(X)
    F5(X)
    F6(X)
    F7(X)
    F8(X)
    F9(X)
    F10(X)
    F11(X)
    F12(X)
    F13(X)
    F14(X)
    F15(X)


def test_CEC2005():
    iters = 10
    pop_size = 10

    for problem in problems_dict.values():
        n_dimension = 2
        sample_size = 100

        left_scalar = problem["bounds"][0]
        right_scalar = problem["bounds"][1]

        X = np.array(
            [np.linspace(left_scalar, right_scalar, sample_size) for _ in range(n_dimension)],
            dtype=np.float64,
        ).T

        function = problem["function"]()

        function.build_grid(x=X[:, 0], y=X[:, 1])

        for dim in problem["dimentions"]:
            print(problem, dim)

            left = np.full(shape=dim, fill_value=left_scalar, dtype=np.float64)
            right = np.full(shape=dim, fill_value=right_scalar, dtype=np.float64)

            optimizer_args = {
                "fitness_function": function,
                "iters": iters,
                "pop_size": pop_size,
                "left": left,
                "right": right,
                "minimization": True,
            }

            if "init_bounds" in problem.keys():
                init_left_scalar = problem["init_bounds"][1]
                init_left = np.full(shape=dim, fill_value=init_left_scalar, dtype=np.float64)
                init_right_scalar = problem["init_bounds"][1]
                init_right = np.full(shape=dim, fill_value=init_right_scalar, dtype=np.float64)

                initial_population = float_population(
                    pop_size=pop_size, left=init_left, right=init_right
                )

                optimizer_args["init_population"] = initial_population

            optimizer = DifferentialEvolution(**optimizer_args)
            optimizer.fit()

    problem = OneMax()

    model = GeneticAlgorithm(fitness_function=problem, iters=iters, pop_size=pop_size, str_len=100)

    model.fit()



from ._datasets import BanknoteDataset
from ._datasets import BreastCancerDataset
from ._datasets import CreditRiskDataset
from ._datasets import DigitsDataset
from ._datasets import IrisDataset
from ._datasets import UserKnowladgeDataset
from ._datasets import WineDataset
from ._optproblems import Ackley
from ._optproblems import Griewank
from ._optproblems import HighConditionedElliptic
from ._optproblems import OneMax
from ._optproblems import Rastrigin
from ._optproblems import Rosenbrock
from ._optproblems import Schwefe1_2
from ._optproblems import Sphere
from ._optproblems import Weierstrass


__all__ = [
    "BanknoteDataset",
    "BreastCancerDataset",
    "CreditRiskDataset",
    "DigitsDataset",
    "IrisDataset",
    "UserKnowladgeDataset",
    "WineDataset",
    "Ackley",
    "Griewank",
    "HighConditionedElliptic",
    "OneMax",
    "Rastrigin",
    "Rosenbrock",
    "Schwefe1_2",
    "Sphere",
    "Weierstrass",
]



import numpy as np

from ..base import EphemeralNode
from ..base import FunctionalNode
from ..base import TerminalNode
from ..base import UniversalSet
from ..optimizers import DifferentialEvolution
from ..optimizers import GeneticProgramming
from ..optimizers import SelfCGA
from ..optimizers import SelfCGP
from ..regressors import GeneticProgrammingNeuralNetRegressor
from ..regressors import MLPEARegressor
from ..regressors import SymbolicRegressionGP
from ..tools.operators import Add
from ..tools.operators import Div
from ..tools.operators import Mul
from ..tools.operators import Neg
from ..tools.operators import uniform_crossoverGP


def test_SymbolicRegressionGP():
    def problem(x):
        return np.sin(x[:, 0])

    def generator1():
        return np.round(np.random.uniform(0, 10), 4)

    def generator2():
        return np.random.randint(0, 10)

    iters = 10
    pop_size = 50

    function = problem
    left_border = -4.5
    right_border = 4.5
    sample_size = 300
    n_dimension = 1

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)]
    ).T
    y = function(X)

    functional_set = [
        FunctionalNode(Add()),
        FunctionalNode(Mul()),
        FunctionalNode(Neg()),
        FunctionalNode(Div()),
    ]

    terminal_set = [TerminalNode(X[:, i], f"x{i}") for i in range(n_dimension)]
    terminal_set.extend([EphemeralNode(generator1), EphemeralNode(generator2)])
    uniset = UniversalSet(functional_set, terminal_set)

    optimizer = GeneticProgramming

    optimizer_args = {
        "tour_size": 15,
        "show_progress_each": 10,
        "optimal_value": 1.1,
        "crossover": "gp_uniform_k",
        "parents_num": 6,
    }

    model = SymbolicRegressionGP(
        iters=iters,
        pop_size=pop_size,
        uniset=uniset,
        optimizer=optimizer,
        optimizer_args=optimizer_args,
    )

    model.fit(X, y)

    optimizer = model.get_optimizers()[0]

    assert optimizer._iters == iters
    assert optimizer._pop_size == pop_size
    assert optimizer._show_progress_each == 10
    assert optimizer._crossover_pool[optimizer._specified_crossover][1] == 6
    assert optimizer._crossover_pool[optimizer._specified_crossover][0] == uniform_crossoverGP

    model = SymbolicRegressionGP(
        iters=iters,
        pop_size=pop_size,
    )

    model.fit(X, y)

    model.predict(X)


def test_GeneticProgrammingNeuralNetRegressor():
    def problem(x):
        return np.sin(x[:, 0])

    iters = 10
    pop_size = 50

    function = problem
    left_border = -4.5
    right_border = 4.5
    sample_size = 300
    n_dimension = 1

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)]
    ).T
    y = function(X)

    optimizer = GeneticProgramming

    optimizer_args = {"tour_size": 15, "show_progress_each": 1}

    iters = 3
    pop_size = 10

    weights_optimizer = DifferentialEvolution
    weights_optimizer_args = {"iters": 25, "pop_size": 25, "CR": 0.9}

    model = GeneticProgrammingNeuralNetRegressor(
        iters=iters,
        pop_size=pop_size,
        optimizer=optimizer,
        optimizer_args=optimizer_args,
        weights_optimizer=weights_optimizer,
        weights_optimizer_args=weights_optimizer_args,
    )

    model.fit(X, y)

    model.predict(X)

    weights_optimizer = SelfCGA
    optimizer = SelfCGP
    weights_optimizer_args = {"iters": 25, "pop_size": 25, "K": 0.33}

    model = GeneticProgrammingNeuralNetRegressor(
        iters=iters,
        pop_size=pop_size,
        optimizer=optimizer,
        optimizer_args=optimizer_args,
        weights_optimizer=weights_optimizer,
        weights_optimizer_args=weights_optimizer_args,
    )

    model.fit(X, y)

    model.predict(X)

    model = GeneticProgrammingNeuralNetRegressor(
        iters=iters,
        pop_size=pop_size,
    )

    model.fit(X, y)

    model.predict(X)

    optimizer = model.get_optimizers()[0]

    assert isinstance(optimizer, model._optimizer_class)


def test_MLPEARegressor():
    def problem(x):
        return np.sin(x[:, 0])

    iters = 10
    pop_size = 50

    function = problem
    left_border = -4.5
    right_border = 4.5
    sample_size = 300
    n_dimension = 1

    X = np.array(
        [np.linspace(left_border, right_border, sample_size) for _ in range(n_dimension)]
    ).T
    y = function(X)

    iters = 3
    pop_size = 10

    weights_optimizer = DifferentialEvolution
    weights_optimizer_args = {"CR": 0.9}

    model = MLPEARegressor(
        iters=iters,
        pop_size=pop_size,
        hidden_layers=(10, 3),
        weights_optimizer=weights_optimizer,
        weights_optimizer_args=weights_optimizer_args,
    )

    model.fit(X, y)

    model.predict(X)

    weights_optimizer = SelfCGA
    weights_optimizer_args = {"K": 0.33}

    model = MLPEARegressor(
        iters=iters,
        pop_size=pop_size,
        hidden_layers=(10,),
        weights_optimizer=weights_optimizer,
        weights_optimizer_args=weights_optimizer_args,
    )

    model.fit(X, y)

    model.predict(X)

    model = MLPEARegressor(iters=iters, pop_size=pop_size, hidden_layers=(10,))

    model.fit(X, y)

    model.predict(X)

    optimizer = model.get_optimizers()[0]

    assert isinstance(optimizer, model._weights_optimizer_class)



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple

import numpy as np
from numpy.typing import NDArray

from ._geneticprogramming import GeneticProgramming
from ._selfcga import SelfCGA
from ..base import UniversalSet
from ..tools import donothing


class SelfCGP(GeneticProgramming, SelfCGA):
    """Semenkin, Eugene & Semenkina, Maria. (2012). Self-configuring genetic programming
    algorithm with modified uniform crossover. 1-6. http://dx.doi.org/10.1109/CEC.2012.6256587"""

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        uniset: UniversalSet,
        iters: int,
        pop_size: int,
        tour_size: int = 2,
        mutation_rate: float = 0.05,
        parents_num: int = 2,
        elitism: bool = True,
        selections: Tuple[str, ...] = (
            "proportional",
            "rank",
            "tournament_3",
            "tournament_5",
            "tournament_7",
        ),
        crossovers: Tuple[str, ...] = ("gp_standart", "gp_one_point", "gp_uniform_rank_2"),
        mutations: Tuple[str, ...] = (
            "gp_weak_point",
            "gp_average_point",
            "gp_strong_point",
            "gp_weak_grow",
            "gp_average_grow",
            "gp_strong_grow",
        ),
        max_level: int = 16,
        init_level: int = 4,
        init_population: Optional[NDArray] = None,
        K: float = 2,
        threshold: float = 0.05,
        genotype_to_phenotype: Callable[[NDArray], NDArray[Any]] = donothing,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
    ):
        SelfCGA.__init__(
            self,
            fitness_function=fitness_function,
            iters=iters,
            pop_size=pop_size,
            str_len=1,
            tour_size=tour_size,
            mutation_rate=mutation_rate,
            parents_num=parents_num,
            elitism=elitism,
            selections=selections,
            crossovers=crossovers,
            mutations=mutations,
            init_population=init_population,
            K=K,
            threshold=threshold,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
        )

        GeneticProgramming.__init__(
            self,
            fitness_function=fitness_function,
            uniset=uniset,
            iters=iters,
            pop_size=pop_size,
            tour_size=tour_size,
            mutation_rate=mutation_rate,
            parents_num=parents_num,
            elitism=elitism,
            max_level=max_level,
            init_level=init_level,
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
        )



import random
from typing import List
from typing import Tuple
from typing import Union

from numba import boolean
from numba import float64
from numba import int64
from numba import njit

import numpy as np
from numpy.typing import NDArray

from ..base import EphemeralNode
from ..base import FunctionalNode
from ..base import TerminalNode
from ..base import Tree
from ..base import UniversalSet
from ..tools import binary_search_interval
from ..tools import check_for_value
from ..tools.transformations import numpy_group_by


def sattolo_shuffle(items: Union[List, NDArray]) -> None:
    i = len(items)
    while i > 1:
        i = i - 1
        j = random.randrange(i)
        items[j], items[i] = items[i], items[j]


def binary_string_population(pop_size: int, str_len: int) -> NDArray[np.byte]:
    size = (pop_size, str_len)
    return np.random.randint(low=2, size=size, dtype=np.byte)


def float_population(
    pop_size: int, left: NDArray[np.float64], right: NDArray[np.float64]
) -> NDArray[np.float64]:
    return np.array(
        [np.random.uniform(left_i, right_i, pop_size) for left_i, right_i in zip(left, right)]
    ).T


@njit(float64[:](float64, float64, int64))
def cauchy_distribution(loc: np.float64, scale: np.float64, size: np.int64) -> NDArray[np.float64]:
    x_ = np.random.standard_cauchy(size=size).astype(np.float64)
    return loc + scale * x_


@njit(float64(float64))
def randc01(u: np.float64) -> np.float64:
    value = cauchy_distribution(loc=u, scale=np.float64(0.1), size=np.int64(1))[0]
    while value <= 0:
        value = cauchy_distribution(loc=u, scale=np.float64(0.1), size=np.int64(1))[0]
    if value > 1:
        value = 1
    return value


@njit(float64(float64))
def randn01(u: np.float64) -> Union[float, np.float64]:
    value = np.random.normal(u, 0.1, size=1)[0]
    if value < 0:
        return 0.0
    elif value > 1:
        return 1.0
    return value


def full_growing_method(uniset: UniversalSet, max_level: int) -> Tree:
    nodes: List[Union[FunctionalNode, TerminalNode, EphemeralNode]] = []
    levels = []
    n_args = []
    possible_steps = [1]
    previous_levels = [-1]
    level_i = -1
    while len(possible_steps):
        possible_steps[-1] = possible_steps[-1] - 1
        if possible_steps[-1] == 0:
            possible_steps.pop()
            level_i = previous_levels.pop() + 1
        else:
            level_i = previous_levels[-1] + 1
        levels.append(level_i)
        if level_i == max_level:
            nodes.append(uniset._random_terminal_or_ephemeral())
            n_args.append(0)
        else:
            nodes.append(uniset._random_functional())
            n_i = nodes[-1]._n_args
            n_args.append(n_i)
            possible_steps.append(n_i)
            previous_levels.append(level_i)
    to_return = Tree(nodes, n_args)
    return to_return


def growing_method(uniset: UniversalSet, max_level: int) -> Tree:
    nodes: List[Union[FunctionalNode, TerminalNode, EphemeralNode]] = []
    levels = []
    n_args = []
    possible_steps = [1]
    previous_levels = [-1]
    level_i = -1
    while len(possible_steps):
        possible_steps[-1] = possible_steps[-1] - 1
        if possible_steps[-1] == 0:
            possible_steps.pop()
            level_i = previous_levels.pop() + 1
        else:
            level_i = previous_levels[-1] + 1
        levels.append(level_i)

        if level_i == max_level:
            nodes.append(uniset._random_terminal_or_ephemeral())
            n_args.append(0)
        elif level_i == 0:
            nodes.append(uniset._random_functional())
            n_i = nodes[-1]._n_args
            n_args.append(n_i)
            possible_steps.append(n_i)
            previous_levels.append(level_i)
        else:
            if np.random.random() < 0.5:
                nodes.append(uniset._random_terminal_or_ephemeral())
            else:
                nodes.append(uniset._random_functional())
            n_i = nodes[-1]._n_args
            n_args.append(n_i)

            if n_i > 0:
                possible_steps.append(n_i)
                previous_levels.append(level_i)
    to_return = Tree(nodes, n_args)
    return to_return


def random_tree(uniset: UniversalSet, max_level: int) -> Tree:
    if random.random() < 0.5:
        to_return = full_growing_method(uniset, max_level)
    else:
        to_return = growing_method(uniset, max_level)
    return to_return


def half_and_half(pop_size: int, uniset: UniversalSet, max_level: int) -> NDArray:
    population = [random_tree(uniset, random.randrange(2, max_level)) for _ in range(pop_size)]
    return np.array(population, dtype=object)


@njit(int64[:](float64[:], int64, boolean))
def random_weighted_sample(
    weights: NDArray[np.float64], quantity: Union[np.int64, int] = 1, replace: bool = True
) -> NDArray[np.int64]:
    if not replace:
        assert len(weights) >= quantity
    to_return = np.empty(quantity, dtype=np.int64)

    cumsumweights = np.cumsum(weights)
    sumweights = cumsumweights[-1]

    i = 0
    while i < quantity:
        roll = sumweights * np.random.rand()
        ind = binary_search_interval(roll, cumsumweights)
        if not replace:
            if check_for_value(ind, to_return, i):
                continue

        to_return[i] = ind
        i += 1
    return to_return


@njit(int64[:](int64, int64, boolean))
def random_sample(
    range_size: np.int64, quantity: np.int64, replace: bool = True
) -> NDArray[np.int64]:
    if not replace:
        assert range_size >= quantity
    to_return = np.empty(quantity, dtype=np.int64)
    i = 0
    while i < quantity:
        ind = random.randrange(range_size)

        if not replace:
            if check_for_value(ind, to_return, i):
                continue

        to_return[i] = ind
        i += 1
    return to_return


def stratified_sample(data: NDArray[np.int64], sample_ratio: float) -> NDArray[np.int64]:
    to_return = []
    data_size = len(data)
    sample_size = int(sample_ratio * data_size)
    indexes = np.arange(len(data), dtype=np.int64)
    keys, groups = numpy_group_by(indexes, data)
    assert sample_size >= len(keys)

    for group in groups:
        group_size = len(group)
        sample_size_i = int((group_size / data_size) * sample_size)
        sample_i_id = random_sample(group_size, sample_size_i, False)
        sample_i = group[sample_i_id]
        to_return.extend(sample_i)

    return np.array(to_return, dtype=np.int64)


def train_test_split_stratified(
    X: NDArray[Union[np.float64, np.int64]],
    y: NDArray[np.int64],
    tests_size: float,
) -> Tuple:
    indexes = np.arange(len(y), dtype=np.int64)
    sample_id = stratified_sample(y, tests_size)
    test_id = sample_id
    train_id = np.setdiff1d(indexes, test_id)
    return (
        X[train_id].astype(np.float64),
        X[test_id].astype(np.float64),
        y[train_id].astype(np.int64),
        y[test_id].astype(np.int64),
    )


def train_test_split(
    X: NDArray[Union[np.float64, np.int64]],
    y: NDArray[Union[np.float64, np.int64]],
    tests_size: float,
) -> Tuple:
    data_size = len(X)
    sample_size = int(tests_size * data_size)
    indexes = np.arange(data_size, dtype=np.int64)
    sample_id = random_sample(range_size=data_size, quantity=sample_size, replace=False)
    test_id = sample_id
    train_id = np.setdiff1d(indexes, test_id)
    return (
        X[train_id].astype(np.float64),
        X[test_id].astype(np.float64),
        y[train_id].astype(np.float64),
        y[test_id].astype(np.float64),
    )



from __future__ import annotations

import random
from typing import Any
from typing import Union

from numba import float64
from numba import int64
from numba import int8
from numba import njit
from numba.types import List as numbaListType

import numpy as np
from numpy.typing import NDArray

from ._numba_funcs import max_axis
from .random import growing_method
from .random import random_sample
from .random import random_weighted_sample
from .random import sattolo_shuffle
from .transformations import common_region
from ..base import EphemeralNode
from ..base import FunctionalNode
from ..base import TerminalNode
from ..base import Tree
from ..base import UniversalSet
from ..base._tree import Operator

min_value = np.finfo(np.float64).min
max_value = np.finfo(np.float64).max


# MUTATUIONS
# genetic algorithm
@njit(int8[:](int8[:], float64))
def flip_mutation(individ: NDArray[np.byte], proba: float) -> NDArray[np.byte]:
    offspring = individ.copy()
    for i in range(offspring.size):
        if random.random() < proba:
            offspring[i] = 1 - offspring[i]
    return offspring


@njit(float64[:](float64[:], float64[:], float64[:, :], float64))
def best_1(
    current: NDArray[np.float64],
    best: NDArray[np.float64],
    population: NDArray[np.float64],
    F: np.float64,
) -> NDArray[np.float64]:
    size = np.int64(len(population))
    r1, r2 = random_sample(range_size=size, quantity=np.int64(2), replace=False)
    return best + F * (population[r1] - population[r2])


@njit(float64[:](float64[:], float64[:], float64[:, :], float64))
def rand_1(
    current: NDArray[np.float64],
    best: NDArray[np.float64],
    population: NDArray[np.float64],
    F: np.float64,
) -> NDArray[np.float64]:
    size = np.int64(len(population))
    r1, r2, r3 = random_sample(range_size=size, quantity=np.int64(3), replace=False)
    return population[r3] + F * (population[r1] - population[r2])


@njit(float64[:](float64[:], float64[:], float64[:, :], float64))
def rand_to_best1(
    current: NDArray[np.float64],
    best: NDArray[np.float64],
    population: NDArray[np.float64],
    F: np.float64,
) -> NDArray[np.float64]:
    size = np.int64(len(population))
    r1, r2, r3 = random_sample(range_size=size, quantity=np.int64(3), replace=False)
    return population[r1] + F * (population[r1]) + F * (population[r2] - population[r3])


@njit(float64[:](float64[:], float64[:], float64[:, :], float64))
def current_to_best_1(
    current: NDArray[np.float64],
    best: NDArray[np.float64],
    population: NDArray[np.float64],
    F: np.float64,
) -> NDArray[np.float64]:
    size = np.int64(len(population))
    r1, r2 = random_sample(range_size=size, quantity=np.int64(2), replace=False)
    return current + F * (best - current) + F * (population[r1] - population[r2])


@njit(float64[:](float64[:], float64[:], float64[:, :], float64))
def best_2(
    current: NDArray[np.float64],
    best: NDArray[np.float64],
    population: NDArray[np.float64],
    F: np.float64,
) -> NDArray[np.float64]:
    size = np.int64(len(population))
    r1, r2, r3, r4 = random_sample(range_size=size, quantity=np.int64(4), replace=False)
    return best + F * (population[r1] - population[r2]) + F * (population[r3] - population[r4])


@njit(float64[:](float64[:], float64[:], float64[:, :], float64))
def rand_2(
    current: NDArray[np.float64],
    best: NDArray[np.float64],
    population: NDArray[np.float64],
    F: np.float64,
) -> NDArray[np.float64]:
    size = np.int64(len(population))
    r1, r2, r3, r4, r5 = random_sample(range_size=size, quantity=np.int64(5), replace=False)
    return (
        population[r5]
        + F * (population[r1] - population[r2])
        + F * (population[r3] - population[r4])
    )


@njit(float64[:](float64[:], float64[:], float64[:, :], float64))
def current_to_rand_1_(
    current: NDArray[np.float64],
    best: NDArray[np.float64],
    population: NDArray[np.float64],
    F: np.float64,
) -> NDArray[np.float64]:
    size = np.int64(len(population))
    r1, r2, r3 = random_sample(range_size=size, quantity=np.int64(3), replace=False)
    return population[r1] + F * (population[r3] - current) + F * (population[r1] - population[r2])


@njit(float64[:](float64[:], float64[:, :], int64[:], float64, float64[:, :]))
def current_to_pbest_1_archive(
    current: NDArray[np.float64],
    population: NDArray[np.float64],
    pbest: NDArray[np.int64],
    F: np.float64,
    pop_archive: NDArray[np.float64],
) -> NDArray[np.float64]:
    p_best_ind = random.randrange(len(pbest))
    best = population[pbest[p_best_ind]]
    r1 = random_sample(range_size=len(population), quantity=np.int64(1), replace=True)[0]
    r2 = random_sample(range_size=len(pop_archive), quantity=np.int64(1), replace=True)[0]
    return current + F * (best - current) + F * (population[r1] - pop_archive[r2])


@njit(float64[:](float64[:], float64[:, :], int64[:], float64, float64[:, :]))
def current_to_pbest_1_archive_p_min(
    current: NDArray[np.float64],
    population: NDArray[np.float64],
    pbest: NDArray[np.int64],
    F: np.float64,
    pop_archive: NDArray[np.float64],
) -> NDArray[np.float64]:
    size = len(population)
    p_min = 2 / size
    p_i = np.random.uniform(p_min, 0.2)
    value = np.int64(p_i * size)
    pbest_cut = pbest[:value]

    p_best_ind = random.randrange(len(pbest_cut))
    best = population[pbest_cut[p_best_ind]]
    r1 = random_sample(range_size=size, quantity=np.int64(1), replace=True)[0]
    r2 = random_sample(range_size=len(pop_archive), quantity=np.int64(1), replace=True)[0]
    return current + F * (best - current) + F * (population[r1] - pop_archive[r2])


# genetic propramming
def point_mutation(tree: Tree, uniset: UniversalSet, proba: float, max_level: int) -> Tree:
    new_node: Union[FunctionalNode, TerminalNode, EphemeralNode]

    to_return = tree.copy()
    if random.random() < proba:
        i = random.randrange(len(to_return))
        if isinstance(to_return._nodes[i], FunctionalNode):
            n_args = to_return._nodes[i]._n_args
            new_node = uniset._random_functional(n_args)
        else:
            new_node = uniset._random_terminal_or_ephemeral()
        to_return._nodes[i] = new_node
    return to_return


def growing_mutation(tree: Tree, uniset: UniversalSet, proba: float, max_level: int) -> Tree:
    to_return = tree.copy()
    if random.random() < proba:
        i = random.randrange(len(to_return))
        grown_tree = growing_method(uniset, max(to_return.get_levels(i)))
        to_return = to_return.concat(i, grown_tree)
    return to_return


def swap_mutation(tree: Tree, uniset: UniversalSet, proba: float, max_leve: int) -> Tree:
    to_return = tree.copy()
    if random.random() < proba:
        more_one_args_cond = to_return._n_args > 1
        indexes = np.arange(len(tree), dtype=int)[more_one_args_cond]
        if len(indexes) > 0:
            i = random.choice(indexes)
            args_id = to_return.get_args_id(i)
            new_arg_id = args_id.copy()
            sattolo_shuffle(new_arg_id)
            for old_j, new_j in zip(args_id, new_arg_id):
                subtree = tree.subtree(old_j)
                to_return = to_return.concat(new_j, subtree)
    return to_return


def shrink_mutation(tree: Tree, uniset: UniversalSet, proba: float, max_level: int) -> Tree:
    to_return = tree.copy()
    if len(to_return) > 2:
        if random.random() < proba:
            no_terminal_cond = to_return._n_args > 0
            indexes = np.arange(len(tree), dtype=int)[no_terminal_cond]
            if len(indexes) > 0:
                i = random.choice(indexes)
                args_id = to_return.get_args_id(i)
                choosen_id = random.choice(args_id)
                to_return = to_return.concat(i, tree.subtree(choosen_id))
    return to_return


# CROSSOVERS
# genetic algorithm
def empty_crossover(
    individs: np.ndarray, fitness: np.ndarray, rank: np.ndarray, *args: Any
) -> np.ndarray:
    offspring = individs[0].copy()
    return offspring


@njit(int8[:](int8[:], int8[:], float64))
def binomialGA(
    individ: NDArray[np.int8], mutant: NDArray[np.int8], CR: np.float64
) -> NDArray[np.byte]:
    size = len(individ)
    offspring = individ.copy()
    j = random.randrange(size)

    for i in range(size):
        if np.random.rand() <= CR or i == j:
            offspring[i] = mutant[i]
    return offspring


@njit(int8[:](int8[:, :], float64[:], float64[:]))
def one_point_crossover(
    individs: NDArray[np.byte], fitness: NDArray[np.float64], rank: NDArray[np.float64]
) -> NDArray[np.byte]:
    cross_point = random_sample(range_size=len(individs[0]), quantity=1, replace=True)[0]
    if random.random() > 0.5:
        offspring = individs[0].copy()
        for i in range(individs.shape[1]):
            if i > cross_point:
                offspring[i] = individs[1][i]
    else:
        offspring = individs[1].copy()
        for i in range(individs.shape[1]):
            if i > cross_point:
                offspring[i] = individs[0][i]
    return offspring


@njit(int8[:](int8[:, :], float64[:], float64[:]))
def two_point_crossover(
    individs: NDArray[np.byte], fitness: NDArray[np.float64], rank: NDArray[np.float64]
) -> NDArray[np.byte]:
    size = len(individs[0])
    c_points = random_sample(range_size=len(individs[0]), quantity=2, replace=False)
    c_points = sorted(c_points)

    if random.random() > 0.5:
        offspring = individs[0].copy()
        other_individ = individs[1]
    else:
        offspring = individs[1].copy()
        other_individ = individs[0]
    for i in range(size):
        if c_points[0] <= i <= c_points[1]:
            offspring[i] = other_individ[i]
    return offspring


@njit(int8[:](int8[:, :], float64[:], float64[:]))
def uniform_crossover(
    individs: NDArray[np.byte], fitness: NDArray[np.float64], rank: NDArray[np.float64]
) -> NDArray[np.byte]:
    choosen = random_sample(range_size=len(fitness), quantity=len(individs[0]), replace=True)
    offspring = np.empty_like(individs[0])
    for i in range(individs.shape[1]):
        offspring[i] = individs[choosen[i]][i]
    return offspring


@njit(int8[:](int8[:, :], float64[:], float64[:]))
def uniform_prop_crossover(
    individs: NDArray[np.byte], fitness: NDArray[np.float64], rank: NDArray[np.float64]
) -> NDArray[np.byte]:
    choosen = random_weighted_sample(weights=fitness, quantity=len(individs[0]), replace=True)
    offspring = np.empty_like(individs[0])
    for i in range(individs.shape[1]):
        offspring[i] = individs[choosen[i]][i]
    return offspring


@njit(int8[:](int8[:, :], float64[:], float64[:]))
def uniform_rank_crossover(
    individs: NDArray[np.byte], fitness: NDArray[np.float64], rank: NDArray[np.float64]
) -> NDArray[np.byte]:
    choosen = random_weighted_sample(weights=rank, quantity=len(individs[0]), replace=True)
    offspring = np.empty_like(individs[0])
    for i in range(individs.shape[1]):
        offspring[i] = individs[choosen[i]][i]
    return offspring


def uniform_tour_crossover(
    individs: NDArray[np.byte], fitness: NDArray[np.float64], rank: NDArray[np.float64]
) -> NDArray[np.byte]:
    range_ = np.arange(len(individs))
    diag = np.arange(len(individs[0]))

    tournament = np.random.choice(range_, 2 * len(individs[0]))
    tournament = tournament.reshape(-1, 2)
    choosen = np.argmax(fitness[tournament], axis=1)
    offspring = individs[choosen, diag].copy()
    return offspring


# differential evolution
@njit(float64[:](float64[:], float64[:], float64))
def binomial(
    individ: NDArray[np.float64], mutant: NDArray[np.float64], CR: np.float64
) -> NDArray[np.float64]:
    size = len(individ)
    offspring = individ.copy()
    j = random.randrange(size)

    for i in range(size):
        if np.random.rand() <= CR or i == j:
            offspring[i] = mutant[i]
    return offspring


# genetic propramming
def standart_crossover(
    individs: NDArray, fitness: NDArray[np.float64], rank: NDArray[np.float64], max_level: int
) -> Tree:
    individ_1 = individs[0].copy()
    individ_2 = individs[1].copy()
    first_point = random.randrange(len(individ_1))
    second_point = random.randrange(len(individ_2))

    if random.random() < 0.5:
        first_subtree = individ_1.subtree(first_point)
        offspring = individ_2.concat(second_point, first_subtree)
        if offspring.get_max_level() > max_level:
            offspring = individ_2
    else:
        second_subtree = individ_2.subtree(second_point)
        offspring = individ_1.concat(first_point, second_subtree)
        if offspring.get_max_level() > max_level:
            offspring = individ_1
    return offspring


def one_point_crossoverGP(
    individs: NDArray, fitness: NDArray[np.float64], rank: NDArray[np.float64], max_level: int
) -> Tree:
    individ_1 = individs[0]
    individ_2 = individs[1]
    common_indexes, _ = common_region([individ_1, individ_2])

    point = random.randrange(len(common_indexes[0]))
    first_point = common_indexes[0][point]
    second_point = common_indexes[1][point]
    if random.random() < 0.5:
        first_subtree = individ_1.subtree(first_point)
        offspring = individ_2.concat(second_point, first_subtree)
    else:
        second_subtree = individ_2.subtree(second_point)
        offspring = individ_1.concat(first_point, second_subtree)
    return offspring


def uniform_crossoverGP(
    individs: NDArray, fitness: NDArray[np.float64], rank: NDArray[np.float64], max_level: int
) -> Tree:
    """Poli, Riccardo & Langdon, W.. (2001). On the Search
    Properties of Different Crossover Operators in Genetic Programming."""
    to_return = Tree([], [])
    new_n_args = []
    common, border = common_region(individs)
    pool = random_sample(range_size=len(fitness), quantity=len(common[0]), replace=True)
    for i, common_0_i in enumerate(common[0]):
        j = pool[i]
        id_ = common[j][i]
        if common_0_i in border[0]:
            subtree = individs[j].subtree(id_)
            to_return._nodes.extend(subtree._nodes)
            new_n_args.extend(subtree._n_args)
        else:
            to_return._nodes.append(individs[j]._nodes[id_])
            new_n_args.append(individs[j]._n_args[id_])

    to_return = to_return.copy()
    to_return._n_args = np.array(new_n_args.copy(), dtype=np.int64)
    return to_return


def uniform_prop_crossover_GP(
    individs: NDArray, fitness: NDArray[np.float64], rank: NDArray[np.float64], max_level: int
) -> Tree:
    to_return = Tree([], [])
    new_n_args = []
    common, border = common_region(individs)
    pool = random_weighted_sample(weights=fitness, quantity=len(common[0]), replace=True)
    for i, common_0_i in enumerate(common[0]):
        j = pool[i]
        id_ = common[j][i]
        if common_0_i in border[0]:
            subtree = individs[j].subtree(id_)
            to_return._nodes.extend(subtree._nodes)
            new_n_args.extend(subtree._n_args)
        else:
            to_return._nodes.append(individs[j]._nodes[id_])
            new_n_args.append(individs[j]._n_args[id_])

    to_return = to_return.copy()
    to_return._n_args = np.array(new_n_args.copy(), dtype=np.int64)
    return to_return


def uniform_rank_crossover_GP(
    individs: NDArray, fitness: NDArray[np.float64], rank: NDArray[np.float64], max_level: int
) -> Tree:
    to_return = Tree([], [])
    new_n_args = []
    common, border = common_region(individs)
    pool = random_weighted_sample(weights=rank, quantity=len(common[0]), replace=True)

    for i, common_0_i in enumerate(common[0]):
        j = pool[i]
        id_ = common[j][i]
        if common_0_i in border[0]:
            subtree = individs[j].subtree(id_)
            to_return._nodes.extend(subtree._nodes)
            new_n_args.extend(subtree._n_args)
        else:
            to_return._nodes.append(individs[j]._nodes[id_])
            new_n_args.append(individs[j]._n_args[id_])

    to_return = to_return.copy()
    to_return._n_args = np.array(new_n_args.copy(), dtype=np.int64)
    return to_return


def uniform_tour_crossover_GP(
    individs: NDArray, fitness: NDArray[np.float64], rank: NDArray[np.float64], max_level: int
) -> Tree:
    to_return = Tree([], [])
    new_n_args = []
    common, border = common_region(individs)
    pool = tournament_selection(fitness, rank, 2, len(common[0]))

    for i, common_0_i in enumerate(common[0]):
        j = pool[i]
        id_ = common[j][i]
        if common_0_i in border[0]:
            subtree = individs[j].subtree(id_)
            to_return._nodes.extend(subtree._nodes)
            new_n_args.extend(subtree._n_args)
        else:
            to_return._nodes.append(individs[j]._nodes[id_])
            new_n_args.append(individs[j]._n_args[id_])

    to_return = to_return.copy()
    to_return._n_args = np.array(new_n_args.copy(), dtype=np.int64)
    return to_return


# SELECTIONS
# genetic algorithm
@njit(int64[:](float64[:], float64[:], int64, int64))
def proportional_selection(
    fitness: np.ndarray, rank: np.ndarray, tour_size: int, quantity: int
) -> np.ndarray:
    choosen = random_weighted_sample(weights=fitness, quantity=quantity, replace=True)
    return choosen


@njit(int64[:](float64[:], float64[:], int64, int64))
def rank_selection(
    fitness: np.ndarray, rank: np.ndarray, tour_size: int, quantity: int
) -> np.ndarray:
    choosen = random_weighted_sample(weights=rank, quantity=quantity, replace=True)
    return choosen


@njit(int64[:](float64[:], float64[:], int64, int64))
def tournament_selection(
    fitness: NDArray[np.float64], rank: NDArray[np.float64], tour_size: np.int64, quantity: np.int64
) -> NDArray[np.int64]:
    to_return = np.empty(quantity, dtype=np.int64)
    for i in range(quantity):
        tournament = random_sample(range_size=len(fitness), quantity=tour_size, replace=False)
        argmax = np.argmax(fitness[tournament])
        to_return[i] = tournament[argmax]
    return to_return


# GP MATH
class Cos(Operator):
    def __init__(self) -> None:
        Operator.__init__(self, formula="cos({})", name="cos", sign="cos")

    def __call__(self, x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        result = np.cos(x)
        return result


class Sin(Operator):
    def __init__(self) -> None:
        Operator.__init__(self, formula="sin({})", name="sin", sign="sin")

    def __call__(self, x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        result = np.sin(x)
        return result


class Add(Operator):
    def __init__(self) -> None:
        Operator.__init__(self, formula="({} + {})", name="add", sign="+")

    def __call__(
        self, x: Union[float, NDArray[np.float64]], y: Union[float, NDArray[np.float64]]
    ) -> Union[float, NDArray[np.float64]]:
        result = x + y
        return result


class Sub(Operator):
    def __init__(self) -> None:
        Operator.__init__(self, formula="({} - {})", name="sub", sign="-")

    def __call__(
        self, x: Union[float, np.ndarray], y: Union[float, np.ndarray]
    ) -> Union[float, np.ndarray]:
        result = x - y
        return result


class Neg(Operator):
    def __init__(self) -> None:
        Operator.__init__(self, formula="-{}", name="neg", sign="-")

    def __call__(self, x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        result = -x
        return result


class Mul(Operator):
    def __init__(self) -> None:
        Operator.__init__(self, formula="({} * {})", name="mul", sign="*")

    def __call__(
        self, x: Union[float, np.ndarray], y: Union[float, np.ndarray]
    ) -> Union[float, np.ndarray]:
        result = x * y
        return result


class Pow2(Operator):
    def __init__(self) -> None:
        Operator.__init__(self, formula="({}**2)", name="pow2", sign="**2")

    def __call__(self, x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        result = np.clip(x**2, min_value, max_value)
        return result


class Div(Operator):
    def __init__(self) -> None:
        Operator.__init__(self, formula="({}/{})", name="div", sign="/")

    def __call__(
        self, x: Union[float, np.ndarray], y: Union[float, np.ndarray]
    ) -> Union[float, np.ndarray]:
        result: Union[float, np.ndarray]
        if isinstance(y, np.ndarray):
            result = np.divide(x, y, out=np.ones_like(y, dtype=np.float64), where=y != 0)
        else:
            if y == 0:
                result = 0.0
            else:
                result = x / y
        result = np.clip(result, min_value, max_value)
        return result


class Inv(Operator):
    def __init__(self) -> None:
        Operator.__init__(self, formula="(1/{})", name="Inv", sign="1/")

    def __call__(self, y: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        result: Union[float, np.ndarray]

        if isinstance(y, np.ndarray):
            result = np.divide(1, y, out=np.ones_like(y, dtype=np.float64), where=y != 0)
        else:
            if y == 0:
                result = 1
            else:
                result = 1 / y
        result = np.clip(result, min_value, max_value)
        return result


class LogAbs(Operator):
    def __init__(self) -> None:
        Operator.__init__(self, formula="log(abs({}))", name="log(abs)", sign="log(abs)")

    def __call__(self, y: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        y_ = np.abs(y)
        if isinstance(y_, np.ndarray):
            result = np.log(y_, out=np.ones_like(y_, dtype=np.float64), where=y_ != 0)
        else:
            if y_ == 0:
                result = 1
            else:
                result = np.log(y_)
        return result


class Exp(Operator):
    def __init__(self) -> None:
        Operator.__init__(self, formula="exp({})", name="exp", sign="exp")

    def __call__(self, x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        result = np.clip(np.exp(x), min_value, max_value)
        return result


class SqrtAbs(Operator):
    def __init__(self) -> None:
        Operator.__init__(self, formula="sqrt(abs({}))", name="sqrt(abs)", sign="sqrt(abs)")

    def __call__(self, x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        result = np.sqrt(np.abs(x))
        return result


class Abs(Operator):
    def __init__(self) -> None:
        Operator.__init__(self, formula="abs({})", name="abs()", sign="abs()")

    def __call__(self, x: Union[float, np.ndarray]) -> Union[float, np.ndarray]:
        result = np.abs(x)
        return result


class More(Operator):
    def __init__(self) -> None:
        Operator.__init__(self, formula="({} > {})", name="more", sign=">")

    def __call__(
        self, x: Union[float, NDArray[np.float64]], y: Union[float, NDArray[np.float64]]
    ) -> Union[bool, NDArray[np.bool_]]:
        result = x > y
        return result


# Activation functions
@njit(float64[:, :](float64[:, :]))
def softmax_numba(X: NDArray[np.float64]) -> NDArray[np.float64]:
    exps = np.exp(X - max_axis(X))
    sum_ = np.sum(exps, axis=1)
    for j in range(sum_.shape[0]):
        if sum_[j] == 0:
            sum_[j] = 1
    result = ((exps).T / sum_).T
    return result


@njit(float64[:, :](float64[:, :], int64))
def multiactivation2d(X: NDArray[np.float64], activ_id: np.int64) -> NDArray[np.float64]:
    if activ_id == 0:
        result = 1 / (1 + np.exp(-X))
    elif activ_id == 1:
        result = X * (X > 0)
    elif activ_id == 2:
        result = np.exp(-(X**2))
    elif activ_id == 3:
        result = np.tanh(X)
    elif activ_id == 4:
        result = softmax_numba(X)
    return result


@njit(float64[:, ::1](float64[:], int64[:, :]))
def mask2d(arr: NDArray[np.float64], mask: NDArray[np.int64]) -> NDArray[np.float64]:
    arr_mask = arr[mask.flatten()]
    arr_mask_reshape = arr_mask.reshape(mask.shape)
    return arr_mask_reshape


@njit
def forward(
    X: NDArray[np.float64],
    weights: NDArray[np.float64],
    nodes: NDArray[np.float64],
    from_: numbaListType(NDArray[np.int64]),
    to_: numbaListType(NDArray[np.int64]),
    weights_id: numbaListType(NDArray[np.int64]),
    activs_code: numbaListType(NDArray[np.int64]),
    activs_nodes: numbaListType(numbaListType(NDArray[np.int64])),
) -> NDArray[np.float64]:
    for from_i, to_i, weights_id_i, a_code_i, a_nodes_i in zip(
        from_, to_, weights_id, activs_code, activs_nodes
    ):
        weights_i = mask2d(weights, weights_id_i)
        out = np.dot(nodes[from_i].T, weights_i.T)
        nodes[to_i] = out.T

        for a_code_i_i, a_nodes_i_i in zip(a_code_i, a_nodes_i):
            nodes[a_nodes_i_i] = multiactivation2d(nodes[a_nodes_i_i].T, a_code_i_i).T

    return nodes


@njit
def forward2d(
    X: NDArray[np.float64],
    inputs: NDArray[np.int64],
    n_hiddens: np.int64,
    outputs: NDArray[np.int64],
    from_: numbaListType(NDArray[np.int64]),
    to_: numbaListType(NDArray[np.int64]),
    weights_id: numbaListType(NDArray[np.int64]),
    activs_code: numbaListType(NDArray[np.int64]),
    activs_nodes: numbaListType(numbaListType(NDArray[np.int64])),
    weights: NDArray[np.float64],
) -> NDArray[np.float64]:
    outs = np.empty(shape=(len(weights), X.shape[0], len(outputs)))
    num_nodes = X.shape[1] + n_hiddens + len(outputs)
    shape = (num_nodes, len(X))
    nodes = np.empty(shape, dtype=np.float64)
    nodes[inputs] = X.T[inputs]

    for n in range(outs.shape[0]):
        forward(X, weights[n], nodes, from_, to_, weights_id, activs_code, activs_nodes)

        outs[n] = nodes[outputs].T
    return outs



import numpy as np

from ._optproblems import ShiftedSphere  # 1 Unimodal (5)
from ._optproblems import ShiftedSchwefe1_2  # 2
from ._optproblems import ShiftedRotatedHighConditionedElliptic  # 3
from ._optproblems import ShiftedSchwefe1_2WithNoise  # 4
from ._optproblems import Schwefel2_6  # 5

from ._optproblems import ShiftedRosenbrock  # 6 Multimodal (20) Basic Functions (7)
from ._optproblems import ShiftedRotatedGriewank  # 7
from ._optproblems import ShiftedRotatedAckley  # 8
from ._optproblems import ShiftedRastrigin  # 9
from ._optproblems import ShiftedRotatedRastrigin  # 10
from ._optproblems import ShiftedRotatedWeierstrass  # 11
from ._optproblems import Schwefel2_13  # 12

from ._optproblems import ShiftedExpandedGriewankRosenbrock  # 13 Expanded Functions (2)
from ._optproblems import ShiftedRotatedExpandedScaffes_F6  # 14

from ._optproblems import HybridCompositionFunction1  # 15 Hybrid Composition Functions (11)
from ._optproblems import RotatedVersionHybridCompositionFunction1  # 16
from ._optproblems import RotatedVersionHybridCompositionFunction1Noise  # 17
from ._optproblems import RotatedHybridCompositionFunction  # 18
from ._optproblems import RotatedHybridCompositionFunctionNarrowBasin  # 19
from ._optproblems import RotatedHybridCompositionFunctionOptimalBounds  # 20
from ._optproblems import HybridCompositionFunction3  # 21
from ._optproblems import HybridCompositionFunction3H  # 22
from ._optproblems import NonContinuousHybridCompositionFunction3  # 23
from ._optproblems import HybridCompositionFunction4  # 24
from ._optproblems import HybridCompositionFunction4withoutbounds  # 25


'''Suganthan, Ponnuthurai & Hansen, Nikolaus & Liang, Jing & Deb, Kalyan & Chen, Ying-ping & Auger,
Anne & Tiwari, Santosh. (2005). Problem Definitions and Evaluation Criteria for the CEC 2005 Special
Session on Real-Parameter Optimization. Natural Computing. 341-357'''

problems_dict = {'F1': {'function': ShiftedSphere,
                        'bounds': (-100, 100),
                        'fix_accuracy': 1e-6,
                        'optimum': -450,
                        'dimentions': range(2, 101)},
                 'F2': {'function': ShiftedSchwefe1_2,
                        'bounds': (-100, 100),
                        'fix_accuracy': 1e-6,
                        'optimum': -450,
                        'dimentions': range(2, 101)},
                 'F3': {'function': ShiftedRotatedHighConditionedElliptic,
                        'bounds': (-100, 100),
                        'fix_accuracy': 1e-6,
                        'optimum': -450,
                        'dimentions': (2, 10, 30, 50)},
                 'F4': {'function': ShiftedSchwefe1_2WithNoise,
                        'bounds': (-100, 100),
                        'fix_accuracy': 1e-6,
                        'optimum': -450,
                        'dimentions': range(2, 101)},
                 'F5': {'function': Schwefel2_6,
                        'bounds': (-100, 100),
                        'fix_accuracy': 1e-6,
                        'optimum': -310,
                        'dimentions': range(2, 101)},
                 'F6': {'function': ShiftedRosenbrock,
                        'bounds': (-100, 100),
                        'fix_accuracy': 1e-2,
                        'optimum': 390,
                        'dimentions': range(2, 101)},
                 'F7': {'function': ShiftedRotatedGriewank,
                        'bounds': (-1000, 1000),
                        'fix_accuracy': 1e-2,
                        'optimum': -180,
                        'init_bounds': (0, 600),
                        'dimentions': (2, 10, 30, 50)},
                 'F8': {'function': ShiftedRotatedAckley,
                        'bounds': (-32, 32),
                        'fix_accuracy': 1e-2,
                        'optimum': -140,
                        'dimentions': (2, 10, 30, 50)},
                 'F9': {'function': ShiftedRastrigin,
                        'bounds': (-5, 5),
                        'fix_accuracy': 1e-2,
                        'optimum': -330,
                        'dimentions': range(2, 51)},
                 'F10': {'function': ShiftedRotatedRastrigin,
                         'bounds': (-5, 5),
                         'fix_accuracy': 1e-2,
                         'optimum': -330,
                         'dimentions': (2, 10, 30, 50)},
                 'F11': {'function': ShiftedRotatedWeierstrass,
                         'bounds': (-0.5, 0.5),
                         'fix_accuracy': 1e-2,
                         'optimum': 90,
                         'dimentions': (2, 10, 30, 50)},
                 'F12': {'function': Schwefel2_13,
                         'bounds': (-np.pi, np.pi),
                         'fix_accuracy': 1e-2,
                         'optimum': -460,
                         'dimentions': range(2, 101)},
                 'F13': {'function': ShiftedExpandedGriewankRosenbrock,
                         'bounds': (-3, 1),
                         'fix_accuracy': 1e-2,
                         'optimum': -130,
                         'dimentions': range(2, 101)},
                 'F14': {'function': ShiftedRotatedExpandedScaffes_F6,
                         'bounds': (-100, 100),
                         'fix_accuracy': 1e-2,
                         'optimum': -300,
                         'dimentions': (2, 10, 30, 50)},
                 'F15': {'function': HybridCompositionFunction1,
                         'bounds': (-5, 5),
                         'fix_accuracy': 1e-2,
                         'optimum': 120,
                         'dimentions': (2, 10, 30, 50)},
                 'F16': {'function': RotatedVersionHybridCompositionFunction1,
                         'bounds': (-5, 5),
                         'fix_accuracy': 1e-2,
                         'optimum': 120,
                         'dimentions': (2, 10, 30, 50)},
                 'F17': {'function': RotatedVersionHybridCompositionFunction1Noise,
                         'bounds': (-5, 5),
                         'fix_accuracy': 1e-1,
                         'optimum': 120,
                         'dimentions': (2, 10, 30, 50)},
                 'F18': {'function': RotatedHybridCompositionFunction,
                         'bounds': (-5, 5),
                         'fix_accuracy': 1e-1,
                         'optimum': 10,
                         'dimentions': (2, 10, 30, 50)},
                 'F19': {'function': RotatedHybridCompositionFunctionNarrowBasin,
                         'bounds': (-5, 5),
                         'fix_accuracy': 1e-1,
                         'optimum': 10,
                         'dimentions': (2, 10, 30, 50)},
                 'F20': {'function': RotatedHybridCompositionFunctionOptimalBounds,
                         'bounds': (-5, 5),
                         'fix_accuracy': 1e-1,
                         'optimum': 10,
                         'dimentions': (2, 10, 30, 50)},
                 'F21': {'function': HybridCompositionFunction3,
                         'bounds': (-5, 5),
                         'fix_accuracy': 1e-1,
                         'optimum': 360,
                         'dimentions': (2, 10, 30, 50)},
                 'F22': {'function': HybridCompositionFunction3H,
                         'bounds': (-5, 5),
                         'fix_accuracy': 1e-1,
                         'optimum': 360,
                         'dimentions': (2, 10, 30, 50)},
                 'F23': {'function': NonContinuousHybridCompositionFunction3,
                         'bounds': (-5, 5),
                         'fix_accuracy': 1e-1,
                         'optimum': 360,
                         'dimentions': (2, 10, 30, 50)},
                 'F24': {'function': HybridCompositionFunction4,
                         'bounds': (-5, 5),
                         'fix_accuracy': 1e-1,
                         'optimum': 260,
                         'dimentions': (2, 10, 30, 50)},
                 'F25': {'function': HybridCompositionFunction4withoutbounds,
                         'bounds': (-10, 10),
                         'fix_accuracy': 1e-1,
                         'optimum': 260,
                         'init_bounds': (2, 5),
                         'dimentions': (2, 10, 30, 50)}}



from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ..base import EphemeralNode
from ..base import FunctionalNode
from ..base import TerminalNode
from ..base import Tree
from ..base import UniversalSet
from ..base._model import Model
from ..base._net import ACTIV_NAME_INV
from ..base._net import HiddenBlock
from ..base._net import Net
from ..optimizers import DifferentialEvolution
from ..optimizers import GeneticAlgorithm
from ..optimizers import GeneticProgramming
from ..optimizers import SHADE
from ..optimizers import SHAGA
from ..optimizers import SelfCGA
from ..optimizers import SelfCGP
from ..optimizers import jDE
from ..tools.metrics import categorical_crossentropy3d
from ..tools.operators import Add
from ..tools.operators import More
from ..tools.random import float_population
from ..tools.random import train_test_split_stratified
from ..tools.transformations import GrayCode


weights_type_optimizer_alias = Union[
    Type[DifferentialEvolution],
    Type[jDE],
    Type[SHADE],
    Type[GeneticAlgorithm],
    Type[SelfCGA],
    Type[SHAGA],
]
weights_optimizer_alias = Union[DifferentialEvolution, jDE, SHADE, GeneticAlgorithm, SelfCGA, SHAGA]


class GeneticProgrammingNeuralNetClassifier(Model):
    """(Lipinsky L., Semenkin E., Bulletin of the Siberian State Aerospace University., 3(10),
    22-26 (2006). In Russian);"""

    def __init__(
        self,
        iters: int,
        pop_size: int,
        input_block_size: int = 1,
        max_hidden_block_size: int = 9,
        offset: bool = True,
        output_activation: str = "softmax",
        test_sample_ratio: float = 0.5,
        optimizer: Union[Type[SelfCGP], Type[GeneticProgramming]] = SelfCGP,
        optimizer_args: Optional[dict[str, Any]] = None,
        weights_optimizer: weights_type_optimizer_alias = SHADE,
        weights_optimizer_args: Optional[dict[str, Any]] = None,
        cache: bool = True,
    ):
        Model.__init__(self)
        self._iters: int = iters
        self._pop_size: int = pop_size
        self._input_block_size: int = input_block_size
        self._max_hidden_block_size: int = max_hidden_block_size
        self._offset: bool = offset
        self._output_activation: str = output_activation
        self._test_sample_ratio: float = test_sample_ratio
        self._optimizer_class: Union[Type[SelfCGP], Type[GeneticProgramming]] = optimizer
        self._optimizer_args: Optional[dict[str, Any]] = optimizer_args
        self._weights_optimizer_class: weights_type_optimizer_alias = weights_optimizer
        self._weights_optimizer_args: Optional[dict[str, Any]] = weights_optimizer_args
        self._cache_condition: bool = cache

        self._optimizer: Union[SelfCGP, GeneticProgramming]
        self._cache: Dict = {}

    def _get_uniset(
        self: GeneticProgrammingNeuralNetClassifier, X: NDArray[np.float64]
    ) -> UniversalSet:
        uniset: UniversalSet
        if self._offset:
            n_dimension = X.shape[1] - 1
        else:
            n_dimension = X.shape[1]

        cut_id: NDArray[np.int64] = np.arange(
            self._input_block_size, n_dimension, self._input_block_size, dtype=np.int64
        )
        variables_pool: List = np.split(np.arange(n_dimension), cut_id)

        functional_set = (FunctionalNode(Add()), FunctionalNode(More()))

        def random_hidden_block() -> HiddenBlock:
            return HiddenBlock(self._max_hidden_block_size)

        terminal_set: List[Union[TerminalNode, EphemeralNode]] = [
            TerminalNode(set(variables), "in{}".format(i))
            for i, variables in enumerate(variables_pool)
        ]
        if self._offset:
            terminal_set.append(
                TerminalNode(value={(n_dimension)}, name="in{}".format(len(variables_pool)))
            )
        terminal_set.append(EphemeralNode(random_hidden_block))

        uniset = UniversalSet(functional_set, tuple(terminal_set))
        return uniset

    def _fitness_function(
        self, population: NDArray, X: NDArray[np.float64], targets: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        output3d = np.array([net.forward(X)[0] for net in population], dtype=np.float64)
        fitness = categorical_crossentropy3d(targets, output3d)
        return fitness

    def _genotype_to_phenotype_tree(self, n_variables: int, n_outputs: int, tree: Tree) -> Net:
        pack: Any = []

        n = n_variables
        for node in reversed(tree._nodes):
            args = []
            for _ in range(node._n_args):
                args.append(pack.pop())
            if isinstance(node, FunctionalNode):
                pack.append(node._value(*args))
            else:
                if type(node) is TerminalNode:
                    unit = Net(inputs=node._value)
                else:
                    end = n + node._value._size
                    hidden_id = set(range(n, end))
                    activs = dict(zip(hidden_id, [node._value._activ] * len(hidden_id)))
                    n = end
                    unit = Net(hidden_layers=[hidden_id], activs=activs)
                pack.append(unit)
        end = n + n_outputs
        output_id = set(range(n, end))
        activs = dict(zip(output_id, [ACTIV_NAME_INV[self._output_activation]] * len(output_id)))
        to_return = pack[0] > Net(outputs=output_id, activs=activs)
        to_return = to_return._fix(set(range(n_variables)))

        return to_return

    def _evaluate_nets(
        self,
        weights: NDArray[np.float64],
        net: Net,
        X: NDArray[np.float64],
        targets: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        output3d = net.forward(X, weights)
        error = categorical_crossentropy3d(targets, output3d)
        return error

    def _train_net(
        self,
        net: Net,
        X_train: NDArray[np.float64],
        proba_train: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        if self._weights_optimizer_args is not None:
            for arg in (
                "fitness_function",
                "left",
                "right",
                "str_len",
                "genotype_to_phenotype",
                "minimization",
            ):
                assert (
                    arg not in self._weights_optimizer_args.keys()
                ), f"""Do not set the "{arg}"
              to the "weights_optimizer_args". It is defined automatically"""
            weights_optimizer_args = self._weights_optimizer_args.copy()

        else:
            weights_optimizer_args = {"iters": 100, "pop_size": 100}

        left: NDArray[np.float64] = np.full(
            shape=len(net._weights), fill_value=-10, dtype=np.float64
        )
        right: NDArray[np.float64] = np.full(
            shape=len(net._weights), fill_value=10, dtype=np.float64
        )
        initial_population: Union[NDArray[np.float64], NDArray[np.byte]] = float_population(
            weights_optimizer_args["pop_size"], left, right
        )
        initial_population[0] = net._weights.copy()
        weights_optimizer_args["fitness_function"] = lambda population: self._evaluate_nets(
            population, net, X_train, proba_train
        )
        if self._weights_optimizer_class in (SHADE, DifferentialEvolution, jDE):
            weights_optimizer_args["left"] = left
            weights_optimizer_args["right"] = right
        else:
            parts: NDArray[np.int64] = np.full(
                shape=len(net._weights), fill_value=16, dtype=np.int64
            )
            genotype_to_phenotype = GrayCode(fit_by="parts").fit(left, right, parts)
            weights_optimizer_args["str_len"] = np.sum(parts)
            weights_optimizer_args["genotype_to_phenotype"] = genotype_to_phenotype.transform

        weights_optimizer_args["minimization"] = True
        optimizer = self._weights_optimizer_class(**weights_optimizer_args)
        optimizer.fit()

        phenotype = optimizer.get_fittest()["phenotype"]
        return phenotype

    def _genotype_to_phenotype(
        self: GeneticProgrammingNeuralNetClassifier,
        X_train: NDArray[np.float64],
        proba_train: NDArray[np.float64],
        population_g: NDArray,
        n_outputs: int,
    ) -> NDArray:
        n_variables: int = X_train.shape[1]

        population_ph: NDArray = np.empty(shape=len(population_g), dtype=object)
        need_to_train_cond: NDArray[np.bool_] = np.full_like(
            population_ph, fill_value=False, dtype=bool
        )

        need_to_train_list: List[Net] = []
        need_to_train_list_str: List = []

        for i, individ_g in enumerate(population_g):
            str_tree = str(individ_g)
            if str_tree in self._cache.keys():
                population_ph[i] = self._cache[str_tree].copy()
            else:
                net = self._genotype_to_phenotype_tree(n_variables, n_outputs, individ_g)

                need_to_train_list.append(net)
                need_to_train_list_str.append(str_tree)
                need_to_train_cond[i] = True

        trained_weights = list(
            map(lambda net: self._train_net(net, X_train, proba_train), need_to_train_list)
        )

        for net, weight in zip(need_to_train_list, trained_weights):
            net._weights = weight

        population_ph[need_to_train_cond] = need_to_train_list

        new_cache = dict(zip(need_to_train_list_str, need_to_train_list))
        self._cache = dict(list(self._cache.items()) + list(new_cache.items()))

        return population_ph

    def get_optimizers(self: GeneticProgrammingNeuralNetClassifier) -> Tuple:
        return (self._optimizer,)

    def _fit(
        self: GeneticProgrammingNeuralNetClassifier,
        X: NDArray[np.float64],
        y: NDArray[Union[np.float64, np.int64]],
    ) -> GeneticProgrammingNeuralNetClassifier:
        optimizer_args: dict[str, Any]

        if self._offset:
            X = np.hstack([X.copy(), np.ones((X.shape[0], 1))])

        n_outputs: int = len(set(y))
        eye: NDArray[np.float64] = np.eye(n_outputs, dtype=np.float64)

        X_train, X_test, y_train, y_test = train_test_split_stratified(
            X, y.astype(np.int64), self._test_sample_ratio
        )

        proba_test: NDArray[np.float64] = eye[y_test]
        proba_train: NDArray[np.float64] = eye[y_train]

        if self._optimizer_args is not None:
            assert (
                "iters" not in self._optimizer_args.keys()
                and "pop_size" not in self._optimizer_args.keys()
            ), """Do not set the "iters" or "pop_size" in the "optimizer_args". Instead,
              use the "SymbolicRegressionGP" arguments"""
            for arg in (
                "fitness_function",
                "uniset",
                "minimization",
            ):
                assert (
                    arg not in self._optimizer_args.keys()
                ), f"""Do not set the "{arg}"
                to the "optimizer_args". It is defined automatically"""
            optimizer_args = self._optimizer_args.copy()

        else:
            optimizer_args = {}

        uniset: UniversalSet = self._get_uniset(X)

        optimizer_args["fitness_function"] = lambda population: self._fitness_function(
            population, X_test, proba_test
        )
        optimizer_args["genotype_to_phenotype"] = lambda trees: self._genotype_to_phenotype(
            X_train, proba_train, trees, n_outputs
        )
        optimizer_args["iters"] = self._iters
        optimizer_args["pop_size"] = self._pop_size
        optimizer_args["uniset"] = uniset
        optimizer_args["minimization"] = True

        self._optimizer = self._optimizer_class(**optimizer_args)
        self._optimizer.fit()

        return self

    def _prepare_output(
        self: GeneticProgrammingNeuralNetClassifier, output: NDArray[np.float64]
    ) -> Union[NDArray[np.float64], NDArray[np.int64]]:
        return np.argmax(output, axis=1)

    def _predict(self, X: NDArray[np.float64]) -> NDArray[Union[np.float64, np.int64]]:
        if self._offset:
            X = np.hstack([X, np.ones((X.shape[0], 1))])

        fittest = self._optimizer.get_fittest()

        output = fittest["phenotype"].forward(X)[0]
        return self._prepare_output(output)



from __future__ import annotations

import random
from collections import defaultdict
from itertools import product
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

from numba.typed import List as numbaList

import numpy as np
from numpy.typing import NDArray

from thefittest.tools.transformations import scale_data

from ..tools.operators import forward2d


INPUT_COLOR_CODE = (0.11, 0.67, 0.47, 1)
HIDDEN_COLOR_CODE = (0.0, 0.74, 0.99, 1)
OUTPUT_COLOR_CODE = (0.94, 0.50, 0.50, 1)
ACTIVATION_NAME = {0: "sg", 1: "rl", 2: "gs", 3: "th", 4: "sm"}
ACTIV_NAME_INV = {"sigma": 0, "relu": 1, "gauss": 2, "tanh": 3, "softmax": 4}


class Net:
    def __init__(
        self,
        inputs: Optional[Set] = None,
        hidden_layers: Optional[List] = None,
        outputs: Optional[Set] = None,
        connects: Optional[NDArray[np.int64]] = None,
        weights: Optional[NDArray[np.float64]] = None,
        activs: Optional[Dict[int, int]] = None,
    ):
        self._inputs = inputs or set()
        self._hidden_layers = hidden_layers or []
        self._outputs = outputs or set()
        self._connects = self._set_connects(values=connects)
        self._weights = self._set_weights(values=weights)
        self._activs = activs or {}

        self._numpy_inputs: Optional[NDArray[np.int64]] = None
        self._numpy_outputs: NDArray[np.int64]
        self._n_hiddens: np.int64

        self._numba_from: numbaList[NDArray[np.int64]]
        self._numba_to: numbaList[NDArray[np.int64]]
        self._numba_weights_id: numbaList[NDArray[np.int64]]
        self._numba_activs_code: numbaList[NDArray[np.int64]]
        self._numba_activs_nodes: numbaList[numbaList[NDArray[np.int64]]]

    def __len__(self) -> int:
        return len(self._weights)

    def _set_connects(self, values: Optional[NDArray[np.int64]]) -> NDArray[np.int64]:
        if values is None:
            to_return = np.empty((0, 2), dtype=np.int64)
        else:
            to_return = values
        return to_return

    def _set_weights(self, values: Optional[NDArray[np.float64]]) -> NDArray[np.float64]:
        if values is None:
            to_return = np.empty((0), dtype=np.float64)
        else:
            to_return = values

        return to_return

    def copy(self) -> Net:
        return Net(
            inputs=self._inputs.copy(),
            hidden_layers=self._hidden_layers.copy(),
            outputs=self._outputs.copy(),
            connects=self._connects.copy(),
            weights=self._weights.copy(),
            activs=self._activs.copy(),
        )

    def _assemble_hiddens(self) -> Set[int]:
        if len(self._hidden_layers) > 0:
            return set.union(*self._hidden_layers)
        else:
            return set()

    def _get_connect(
        self, left: Set[int], right: Set[int]
    ) -> Tuple[NDArray[np.int64], NDArray[np.float64]]:
        if len(left) and len(right):
            connects = np.array(list(product(left, right)), dtype=np.int64)
            weights = np.random.uniform(-2, 2, len(connects)).astype(np.float64)
            return (connects, weights)
        else:
            return (np.zeros((0, 2), dtype=np.int64), np.zeros((0), dtype=np.float64))

    def __add__(self, other: Net) -> Net:
        len_i_1, len_i_2 = len(self._inputs), len(other._inputs)
        len_h_1, len_h_2 = len(self._hidden_layers), len(other._hidden_layers)

        if (len_i_1 > 0 and len_i_2 == 0) and (len_h_1 == 0 and len_h_2 > 0):
            return self > other
        elif (len_i_1 == 0 and len_i_2 > 0) and (len_h_1 > 0 and len_h_2 == 0):
            return other > self

        map_res = map(
            lambda layers: layers[0].union(layers[1]),
            zip(self._hidden_layers, other._hidden_layers),
        )

        if len_h_1 < len_h_2:
            excess = other._hidden_layers[len_h_1:]
        elif len_h_1 > len_h_2:
            excess = self._hidden_layers[len_h_2:]
        else:
            excess = []

        hidden = list(map_res) + excess
        return Net(
            inputs=self._inputs.union(other._inputs),
            hidden_layers=hidden,
            outputs=self._outputs.union(other._outputs),
            connects=np.vstack([self._connects, other._connects]),
            weights=np.hstack([self._weights, other._weights]),
            activs={**self._activs, **other._activs},
        )

    def __gt__(self, other: Net) -> Net:
        len_i_1, len_i_2 = len(self._inputs), len(other._inputs)
        len_h_1, len_h_2 = len(self._hidden_layers), len(other._hidden_layers)

        if (len_i_1 > 0 and len_h_1 == 0) and (len_i_2 > 0 and len_h_2 == 0):
            return self + other
        elif (len_i_1 == 0 and len_h_1 > 0) and (len_i_2 > 0 and len_h_2 == 0):
            return other > self

        inputs_hidden = self._inputs.union(self._assemble_hiddens())
        from_ = inputs_hidden.difference(self._connects[:, 0])

        cond = other._connects[:, 0][:, np.newaxis] == np.array(list(other._inputs))
        cond = np.any(cond, axis=1)

        connects_no_i = other._connects[:, 1][~cond]
        hidden_outputs = other._assemble_hiddens().union(other._outputs)
        to_ = hidden_outputs.difference(connects_no_i)

        connects, weights = self._get_connect(from_, to_)
        return Net(
            inputs=self._inputs.union(other._inputs),
            hidden_layers=self._hidden_layers + other._hidden_layers,
            outputs=self._outputs.union(other._outputs),
            connects=np.vstack([self._connects, other._connects, connects]),
            weights=np.hstack([self._weights, other._weights, weights]),
            activs={**self._activs, **other._activs},
        )

    def _fix(self, inputs: Set[int]) -> Net:
        hidden_outputs = self._assemble_hiddens().union(self._outputs)
        to_ = hidden_outputs.difference(self._connects[:, 1])
        if len(to_) > 0:
            if not len(self._inputs):
                self._inputs = inputs

            connects, weights = self._get_connect(self._inputs, to_)
            self._connects = np.vstack([self._connects, connects])
            self._weights = np.hstack([self._weights, weights])

        self._connects = np.unique(self._connects, axis=0)
        self._weights = self._weights[: len(self._connects)]
        return self

    def _get_order(self) -> None:
        hidden = self._assemble_hiddens()
        from_ = self._connects[:, 0]
        to_ = self._connects[:, 1]
        indexes = np.arange(len(from_), dtype=np.int64)

        argsort = np.argsort(to_)

        from_sort = from_[argsort]
        to_sort = to_[argsort]
        index_sort = indexes[argsort]

        groups_to, cut_index = np.unique(to_sort, return_index=True)
        groups_from = np.split(from_sort, cut_index)[1:]
        group_index = np.split(index_sort, cut_index)[1:]

        pairs = defaultdict(list)
        weights_id_list = defaultdict(list)
        weights_id_ndarray = {}

        for groups_to_i, groups_from_i, group_index_i in zip(groups_to, groups_from, group_index):
            argsort = np.argsort(groups_from_i)
            groups_from_i_sort = tuple(groups_from_i[argsort])
            pairs[groups_from_i_sort].append(groups_to_i)
            weights_id_list[groups_from_i_sort].append(group_index_i[argsort])

        for key, value in weights_id_list.items():
            weights_id_ndarray[key] = np.array(value, dtype=np.int64)

        numba_from = numbaList()
        numba_to = numbaList()
        numba_weight_id = numbaList()

        activ_code = numbaList()
        active_nodes = numbaList()

        calculated = self._inputs.copy()
        purpose = self._inputs.union(hidden).union(self._outputs)

        while calculated != purpose:
            for from_i, to_i in pairs.items():
                if set(from_i).issubset(calculated) and not set(to_i).issubset(calculated):
                    calculated = calculated.union(set(to_i))
                    numba_from.append(np.array(from_i, dtype=np.int64))
                    numba_to.append(np.array(to_i, dtype=np.int64))
                    numba_weight_id.append(weights_id_ndarray[from_i])

                    nodes_i = defaultdict(list)
                    for to_i_i in to_i:
                        nodes_i[self._activs[to_i_i]].append(to_i_i)

                    activ_code.append(np.array(list(nodes_i.keys()), dtype=np.int64))

                    nodes_i_list = numbaList()
                    for value in nodes_i.values():
                        nodes_i_list.append(np.array(value, dtype=np.int64))

                    active_nodes.append(nodes_i_list)

        self._numpy_inputs = np.array(list(self._inputs), dtype=np.int64)
        self._numpy_outputs = np.array(list(self._outputs), dtype=np.int64)
        self._n_hiddens = np.int64(len(hidden))
        self._numba_from = numba_from
        self._numba_to = numba_to
        self._numba_weights_id = numba_weight_id
        self._numba_activs_code = activ_code
        self._numba_activs_nodes = active_nodes

    def forward(
        self, X: NDArray[np.float64], weights: Optional[NDArray[np.float64]] = None
    ) -> NDArray[np.float64]:
        if weights is None:
            weights = self._weights.reshape(1, -1)
        else:
            weights = weights

        if self._numpy_inputs is None:
            self._get_order()

        outputs = forward2d(
            X,
            self._numpy_inputs,
            self._n_hiddens,
            self._numpy_outputs,
            self._numba_from,
            self._numba_to,
            self._numba_weights_id,
            self._numba_activs_code,
            self._numba_activs_nodes,
            weights,
        )
        return outputs

    def get_graph(self) -> Dict:
        weights_scale = scale_data(self._weights)
        nodes = list(self._inputs)

        len_i = len(self._inputs)
        len_h = len(self._assemble_hiddens())
        len_o = len(self._outputs)
        sum_ = len_i + len_h + len_o

        positions = np.zeros((sum_, 2), dtype=float)
        colors = np.zeros((sum_, 4))
        w_colors = np.zeros((len(weights_scale), 4))
        labels = {
            **dict(zip(self._inputs, self._inputs)),
            **{key: ACTIVATION_NAME[value] for key, value in self._activs.items()},
            **dict(zip(self._outputs, range(len_o))),
        }

        w_colors[:, 0] = 1 - weights_scale
        w_colors[:, 2] = weights_scale
        w_colors[:, 3] = 0.8
        positions[:len_i][:, 1] = np.arange(len_i) - (len_i) / 2
        colors[:len_i] = INPUT_COLOR_CODE

        n = len_i
        for i, layer in enumerate(self._hidden_layers):
            nodes.extend(list(layer))
            positions[n : n + len(layer)][:, 0] = i + 1
            positions[n : n + len(layer)][:, 1] = np.arange(len(layer)) - len(layer) / 2
            colors[n : n + len(layer)] = HIDDEN_COLOR_CODE
            n += len(layer)

        nodes.extend(list(self._outputs))
        positions[n : n + len_o][:, 0] = len(self._hidden_layers) + 1
        positions[n : n + len_o][:, 1] = np.arange(len_o) - len_o / 2
        colors[n : n + len_o] = OUTPUT_COLOR_CODE

        positions_dict = dict(zip(nodes, positions))

        to_return = {
            "nodes": nodes,
            "labels": labels,
            "positions": positions_dict,
            "colors": colors,
            "weights_colors": w_colors,
            "connects": self._connects,
        }

        return to_return


class HiddenBlock:
    def __init__(self, max_size: int) -> None:
        self._activ = random.sample([0, 1, 2, 3], k=1)[0]
        self._size = random.randrange(1, max_size)

    def __str__(self) -> str:
        return "{}{}".format(ACTIVATION_NAME[self._activ], self._size)



import os
from typing import Dict
from typing import Union

import numpy as np
from numpy.typing import NDArray


path = os.path.dirname(__file__) + "/_data/"

iris_data = np.loadtxt(path + "iris_data.txt", delimiter=",")
wine_data = np.loadtxt(path + "wine_data.txt", delimiter=",")
breast_data = np.loadtxt(path + "breast_cancer_data.txt", delimiter=",")
digits_data = np.loadtxt(path + "handwritten_digits_data.txt", delimiter=",")
credit_data = np.loadtxt(path + "credit_risk_data.txt", delimiter=",")
know_data = np.loadtxt(path + "user_knowledge_data.txt", delimiter=",")
banknote_data = np.loadtxt(path + "banknote_dataset.txt", delimiter=",")


class Dataset:
    def __init__(
        self,
        X: NDArray[np.float64],
        y: NDArray[Union[np.int64, np.float64]],
        X_names: Dict,
        y_names: Dict,
    ):
        self._X = X
        self._y = y
        self._X_names = X_names
        self._y_names = y_names

    def get_X(self) -> NDArray[np.float64]:
        return self._X

    def get_y(self) -> NDArray[Union[np.int64, np.float64]]:
        return self._y

    def get_X_names(self) -> Dict[int, str]:
        return self._X_names

    def get_y_names(self) -> Dict[int, str]:
        return self._y_names


class IrisDataset(Dataset):
    """Fisher R. A.. (1988). Iris. UCI Machine Learning Repository.
    https://doi.org/10.24432/C56C76."""

    def __init__(self) -> None:
        Dataset.__init__(
            self,
            X=iris_data[:, :-1].astype(np.float64),
            y=iris_data[:, -1].astype(np.int64),
            X_names={
                0: "sepal length in cm",
                1: "sepal width in cm",
                2: "petal length in cm",
                3: "petal width in cm",
            },
            y_names={0: "Iris-setosa", 1: "Iris-versicolor", 2: "Iris-virginica"},
        )


class WineDataset(Dataset):
    """Aeberhard,Stefan and Forina,M.. (1991). Wine. UCI Machine
    Learning Repository. https://doi.org/10.24432/C5PC7J."""

    def __init__(self) -> None:
        Dataset.__init__(
            self,
            X=wine_data[:, 1:].astype(np.float64),
            y=wine_data[:, 0].astype(np.int64) - 1,
            X_names={
                0: "Alcohol",
                1: "Malic acid",
                2: "Ash",
                3: "Alcalinity of ash",
                4: "Magnesium",
                5: "Total phenols",
                6: "Flavanoids",
                7: "Nonflavanoid phenols",
                8: "Proanthocyanins",
                9: "Color intensity",
                10: "Hue",
                11: "OD280/OD315 of diluted wines",
                12: "Proline",
            },
            y_names={0: "class 1", 1: "class 2", 2: "class 3"},
        )


class BreastCancerDataset(Dataset):
    """
    Wolberg,William, Mangasarian,Olvi, Street,Nick, and Street,W.. (1995).
    Breast Cancer Wisconsin (Diagnostic). UCI Machine Learning Repository.
    https://doi.org/10.24432/C5DW2B."""

    def __init__(self) -> None:
        Dataset.__init__(
            self,
            X=breast_data[:, 2:].astype(np.float64),
            y=breast_data[:, 1].astype(np.int64),
            X_names={
                0: "mean radius",
                1: "mean texture",
                2: "mean perimeter",
                3: "mean area",
                4: "mean smoothness",
                5: "mean compactness",
                6: "mean concavity",
                7: "mean concave points",
                8: "mean symmetry",
                9: "mean fractal dimension",
                10: "radius error",
                11: "texture error",
                12: "perimeter error",
                13: "area error",
                14: "smoothness error",
                15: "compactness error",
                16: "concavity error",
                17: "concave points error",
                18: "symmetry error",
                19: "fractal dimension error",
                20: "worst radius",
                21: "worst texture",
                22: "worst perimeter",
                23: "worst area",
                24: "worst smoothness",
                25: "worst compactness",
                26: "worst concavity",
                27: "worst concave points",
                28: "worst symmetry",
                29: "worst fractal dimension",
            },
            y_names={0: "M", 1: "B"},
        )


class DigitsDataset(Dataset):
    """Alpaydin,E. and Kaynak,C.. (1998). Optical Recognition of Handwritten Digits.
    UCI Machine Learning Repository. https://doi.org/10.24432/C50P49."""

    def __init__(self) -> None:
        Dataset.__init__(
            self,
            X=digits_data[:, :-1].astype(np.float64),
            y=digits_data[:, -1].astype(np.int64),
            X_names=dict(zip(range(64), range(64))),
            y_names=dict(zip(range(10), range(10))),
        )


class CreditRiskDataset(Dataset):
    """https://www.kaggle.com/datasets/upadorprofzs/credit-risk"""

    def __init__(self) -> None:
        Dataset.__init__(
            self,
            X=credit_data[:, 1:-1].astype(np.float64),
            y=credit_data[:, -1].astype(np.int64),
            X_names={0: "income", 1: "age", 2: "loan"},
            y_names={0: "good client", 1: "bad client"},
        )


class UserKnowladgeDataset(Dataset):
    """Kahraman,Hamdi, Colak,Ilhami, and Sagiroglu,Seref. (2013). User Knowledge Modeling.
    UCI Machine Learning Repository. https://doi.org/10.24432/C5231X."""

    def __init__(self) -> None:
        Dataset.__init__(
            self,
            X=know_data[:, :-1].astype(np.float64),
            y=know_data[:, -1].astype(np.int64),
            X_names={
                0: "STG (The degree of study time for goal object materails)",
                1: "SCG (The degree of repetition number of user for goal object materails) ",
                2: "STR (The degree of study time of user for related objects with goal object)",
                3: "LPR (The exam performance of user for related objects with goal object)",
                4: "PEG (The exam performance of user for goal objects)",
            },
            y_names={0: "Very Low", 1: "Low", 2: "Middle", 3: "High "},
        )


class BanknoteDataset(Dataset):
    """
    Lohweg,Volker. (2013). banknote authentication.
    UCI Machine Learning Repository. https://doi.org/10.24432/C55P57."""

    def __init__(self) -> None:
        Dataset.__init__(
            self,
            X=banknote_data[:, :-1].astype(np.float64),
            y=banknote_data[:, -1].astype(np.int64),
            X_names={
                0: "variance of Wavelet Transformed image (continuous)",
                1: "skewness of Wavelet Transformed image (continuous)",
                2: "curtosis of Wavelet Transformed image (continuous)",
                3: "entropy of image (continuous)",
            },
            y_names={0: "not original", 1: "original"},
        )



from ..benchmarks import IrisDataset
from ..classifiers import GeneticProgrammingNeuralNetClassifier
from ..classifiers import MLPEAClassifier
from ..optimizers import DifferentialEvolution
from ..optimizers import GeneticProgramming
from ..optimizers import SelfCGA
from ..tools.transformations import scale_data


def test_GeneticProgrammingNeuralNetClassifier():
    data = IrisDataset()
    X = data.get_X()
    y = data.get_y()

    data = IrisDataset()
    X = scale_data(data.get_X())
    y = data.get_y()

    optimizer = GeneticProgramming

    optimizer_args = {"tour_size": 15, "show_progress_each": 1}

    iters = 3
    pop_size = 10

    weights_optimizer = DifferentialEvolution
    weights_optimizer_args = {"iters": 25, "pop_size": 25, "CR": 0.9}

    model = GeneticProgrammingNeuralNetClassifier(
        iters=iters,
        pop_size=pop_size,
        optimizer=optimizer,
        optimizer_args=optimizer_args,
        weights_optimizer=weights_optimizer,
        weights_optimizer_args=weights_optimizer_args,
    )

    model.fit(X, y)

    model.predict(X)

    weights_optimizer = SelfCGA
    weights_optimizer_args = {"iters": 25, "pop_size": 25, "K": 0.33}

    model = GeneticProgrammingNeuralNetClassifier(
        iters=iters,
        pop_size=pop_size,
        optimizer=optimizer,
        optimizer_args=optimizer_args,
        weights_optimizer=weights_optimizer,
        weights_optimizer_args=weights_optimizer_args,
    )

    model.fit(X, y)

    model.predict(X)

    model = GeneticProgrammingNeuralNetClassifier(
        iters=iters,
        pop_size=pop_size,
    )

    model.fit(X, y)

    model.predict(X)

    optimizer = model.get_optimizers()[0]

    assert isinstance(optimizer, model._optimizer_class)


def test_MLPEAClassifier():
    data = IrisDataset()
    X = data.get_X()
    y = data.get_y()

    data = IrisDataset()
    X = scale_data(data.get_X())
    y = data.get_y()

    iters = 3
    pop_size = 10

    weights_optimizer = DifferentialEvolution
    weights_optimizer_args = {"CR": 0.9}

    model = MLPEAClassifier(
        iters=iters,
        pop_size=pop_size,
        hidden_layers=(10,),
        weights_optimizer=weights_optimizer,
        weights_optimizer_args=weights_optimizer_args,
    )

    model.fit(X, y)

    model.predict(X)

    weights_optimizer = SelfCGA
    weights_optimizer_args = {"K": 0.33}

    model = MLPEAClassifier(
        iters=iters,
        pop_size=pop_size,
        hidden_layers=(1, 2),
        weights_optimizer=weights_optimizer,
        weights_optimizer_args=weights_optimizer_args,
    )

    model.fit(X, y)

    model.predict(X)

    model = MLPEAClassifier(
        iters=iters,
        pop_size=pop_size,
        hidden_layers=(0,),
    )

    model.fit(X, y)

    model.predict(X)

    optimizer = model.get_optimizers()[0]

    assert isinstance(optimizer, model._weights_optimizer_class)



from ._gpnnregression import GeneticProgrammingNeuralNetRegressor
from ._mlpearegressor import MLPEARegressor
from ._symbolicregressiongp import SymbolicRegressionGP


__all__ = ["SymbolicRegressionGP", "GeneticProgrammingNeuralNetRegressor", "MLPEARegressor"]



from __future__ import annotations

from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ..base import EphemeralNode
from ..base import FunctionalNode
from ..base import TerminalNode
from ..base import Tree
from ..base import UniversalSet
from ..base._model import Model
from ..optimizers import GeneticProgramming
from ..optimizers import SelfCGP
from ..tools.metrics import coefficient_determination
from ..tools.operators import Add
from ..tools.operators import Cos
from ..tools.operators import Inv
from ..tools.operators import Mul
from ..tools.operators import Neg
from ..tools.operators import Pow2
from ..tools.operators import Sin


class SymbolicRegressionGP(Model):
    def __init__(
        self,
        iters: int,
        pop_size: int,
        uniset: Optional[UniversalSet] = None,
        optimizer: Union[Type[SelfCGP], Type[GeneticProgramming]] = SelfCGP,
        optimizer_args: Optional[dict[str, Any]] = None,
    ):
        Model.__init__(self)

        self._iters: int = iters
        self._pop_size: int = pop_size
        self._uniset: Optional[UniversalSet] = uniset
        self._optimizer_args: Optional[dict[str, Any]] = optimizer_args
        self._optimizer_class: Union[Type[SelfCGP], Type[GeneticProgramming]] = optimizer
        self._optimizer: Union[SelfCGP, GeneticProgramming]

    @staticmethod
    def _generator1() -> float:
        value = np.round(np.random.uniform(0, 10), 4)
        return value

    @staticmethod
    def _generator2() -> int:
        value = np.random.randint(0, 10)
        return value

    def _get_uniset(self: SymbolicRegressionGP, X: NDArray[np.float64]) -> UniversalSet:
        uniset: UniversalSet
        terminal_set: Union[
            List[Union[TerminalNode, EphemeralNode]], Tuple[Union[TerminalNode, EphemeralNode]]
        ]
        n_dimension: int = X.shape[1]

        functional_set = (
            FunctionalNode(Add()),
            FunctionalNode(Mul()),
            FunctionalNode(Neg()),
            FunctionalNode(Inv()),
            FunctionalNode(Pow2()),
            FunctionalNode(Cos()),
            FunctionalNode(Sin()),
        )
        terminal_set = [TerminalNode(X[:, i], f"x{i}") for i in range(n_dimension)]
        terminal_set.extend([EphemeralNode(self._generator1), EphemeralNode(self._generator2)])
        uniset = UniversalSet(functional_set, tuple(terminal_set))
        return uniset

    def _evaluate_tree(self: SymbolicRegressionGP, tree: Tree, y: NDArray[np.float64]) -> float:
        y_pred = tree() * np.ones(len(y))
        fitness = coefficient_determination(y, y_pred)
        return fitness

    def _fitness_function(
        self: SymbolicRegressionGP, trees: NDArray, y: NDArray[np.float64]
    ) -> NDArray[np.float64]:
        fitness = [self._evaluate_tree(tree, y) for tree in trees]
        return np.array(fitness)

    def get_optimizers(self: SymbolicRegressionGP) -> Tuple:
        return (self._optimizer,)

    def _fit(
        self: SymbolicRegressionGP, X: NDArray[np.float64], y: NDArray[Union[np.float64, np.int64]]
    ) -> SymbolicRegressionGP:
        optimizer_args: dict[str, Any]
        uniset: UniversalSet

        if self._uniset is None:
            uniset = self._get_uniset(X)
        else:
            uniset = self._uniset

        if self._optimizer_args is not None:
            assert (
                "iters" not in self._optimizer_args.keys()
                and "pop_size" not in self._optimizer_args.keys()
                and "uniset" not in self._optimizer_args.keys()
            ), """Do not set the "iters", "pop_size", or "uniset" in the "optimizer_args". Instead,
              use the "SymbolicRegressionGP" arguments"""
            assert (
                "fitness_function" not in self._optimizer_args.keys()
            ), """Do not set the "fitness_function"
              to the "optimizer_args". It is defined automatically"""
            assert (
                "minimization" not in self._optimizer_args.keys()
            ), """Do not set the "minimization"
              to the "optimizer_args". It is defined automatically"""
            optimizer_args = self._optimizer_args.copy()

        else:
            optimizer_args = {}

        optimizer_args["fitness_function"] = lambda trees: self._fitness_function(
            trees, y.astype(np.float64)
        )
        optimizer_args["iters"] = self._iters
        optimizer_args["pop_size"] = self._pop_size
        optimizer_args["uniset"] = uniset

        self._optimizer = self._optimizer_class(**optimizer_args)
        self._optimizer.fit()

        return self

    def _predict(
        self: SymbolicRegressionGP, X: NDArray[np.float64]
    ) -> NDArray[Union[np.float64, np.int64]]:
        n_dimension = X.shape[1]
        solution = self.get_optimizers()[0].get_fittest()

        genotype_for_pred = solution["phenotype"].set_terminals(
            **{f"x{i}": X[:, i] for i in range(n_dimension)}
        )

        y_pred = genotype_for_pred() * np.ones(len(X))
        return y_pred



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ._geneticalgorithm import GeneticAlgorithm
from ..tools import donothing
from ..tools.transformations import numpy_group_by


class SelfCGA(GeneticAlgorithm):
    """Semenkin, E.S., Semenkina, M.E. Self-configuring Genetic Algorithm with Modified Uniform
    Crossover Operator. LNCS, 7331, 2012, pp. 414-421. https://doi.org/10.1007/978-3-642-30976-2_50
    """

    def __init__(
        self,
        fitness_function: Callable[[NDArray[Any]], NDArray[np.float64]],
        iters: int,
        pop_size: int,
        str_len: int,
        tour_size: int = 2,
        mutation_rate: float = 0.05,
        parents_num: int = 2,
        elitism: bool = True,
        selections: Tuple[str, ...] = (
            "proportional",
            "rank",
            "tournament_3",
            "tournament_5",
            "tournament_7",
        ),
        crossovers: Tuple[str, ...] = (
            "empty",
            "one_point",
            "two_point",
            "uniform_2",
            "uniform_7",
            "uniform_prop_2",
            "uniform_prop_7",
            "uniform_rank_2",
            "uniform_rank_7",
            "uniform_tour_3",
            "uniform_tour_7",
        ),
        mutations: Tuple[str, ...] = ("weak", "average", "strong"),
        init_population: Optional[NDArray[np.byte]] = None,
        K: float = 2,
        threshold: float = 0.05,
        genotype_to_phenotype: Callable[[NDArray[np.byte]], NDArray[Any]] = donothing,
        optimal_value: Optional[float] = None,
        termination_error_value: float = 0.0,
        no_increase_num: Optional[int] = None,
        minimization: bool = False,
        show_progress_each: Optional[int] = None,
        keep_history: bool = False,
    ):
        GeneticAlgorithm.__init__(
            self,
            fitness_function=fitness_function,
            iters=iters,
            pop_size=pop_size,
            str_len=str_len,
            tour_size=tour_size,
            mutation_rate=mutation_rate,
            parents_num=parents_num,
            elitism=elitism,
            init_population=init_population,
            genotype_to_phenotype=genotype_to_phenotype,
            optimal_value=optimal_value,
            termination_error_value=termination_error_value,
            no_increase_num=no_increase_num,
            minimization=minimization,
            show_progress_each=show_progress_each,
            keep_history=keep_history,
        )

        self._K = K
        self._threshold = threshold

        self._selection_set: Dict[str, Tuple[Callable, Union[float, int]]] = {}
        self._crossover_set: Dict[str, Tuple[Callable, Union[float, int]]] = {}
        self._mutation_set: Dict[str, Tuple[Callable, Union[float, int], bool]] = {}

        self._selection_proba: Dict[str, float]
        self._crossover_proba: Dict[str, float]
        self._mutation_proba: Dict[str, float]

        for operator_name in selections:
            self._selection_set[operator_name] = self._selection_pool[operator_name]
        self._selection_set = dict(sorted(self._selection_set.items()))

        for operator_name in crossovers:
            self._crossover_set[operator_name] = self._crossover_pool[operator_name]
        self._crossover_set = dict(sorted(self._crossover_set.items()))

        for operator_name in mutations:
            self._mutation_set[operator_name] = self._mutation_pool[operator_name]
        self._mutation_set = dict(sorted(self._mutation_set.items()))

        self._z_selection = len(self._selection_set)
        self._z_crossover = len(self._crossover_set)
        self._z_mutation = len(self._mutation_set)

        self._selection_proba = dict(
            zip(list(self._selection_set.keys()), np.full(self._z_selection, 1 / self._z_selection))
        )
        if "empty" in self._crossover_set.keys():
            self._crossover_proba = dict(
                zip(
                    list(self._crossover_set.keys()),
                    np.full(self._z_crossover, 0.9 / (self._z_crossover - 1)),
                )
            )
            self._crossover_proba["empty"] = 0.1
        else:
            self._crossover_proba = dict(
                zip(
                    list(self._crossover_set.keys()),
                    np.full(self._z_crossover, 1 / self._z_crossover),
                )
            )
        self._mutation_proba = dict(
            zip(list(self._mutation_set.keys()), np.full(self._z_mutation, 1 / self._z_mutation))
        )

        self._selection_operators: NDArray = self._choice_operators(
            proba_dict=self._selection_proba
        )
        self._crossover_operators: NDArray = self._choice_operators(
            proba_dict=self._crossover_proba
        )
        self._mutation_operators: NDArray = self._choice_operators(proba_dict=self._mutation_proba)

    def _choice_operators(self: SelfCGA, proba_dict: Dict["str", float]) -> NDArray:
        operators = list(proba_dict.keys())
        proba = list(proba_dict.values())
        chosen_operator = np.random.choice(operators, self._pop_size, p=proba)
        return chosen_operator

    def _get_new_proba(
        self: SelfCGA, proba_dict: Dict["str", float], operator: str
    ) -> Dict["str", float]:
        proba_dict[operator] += self._K / self._iters
        proba_value = np.array(list(proba_dict.values()))
        proba_value -= self._K / (len(proba_dict) * self._iters)
        proba_value = proba_value.clip(self._threshold, 1)
        proba_value = proba_value / proba_value.sum()
        new_proba_dict = dict(zip(proba_dict.keys(), proba_value))
        return new_proba_dict

    def _find_fittest_operator(
        self: SelfCGA, operators: NDArray, fitness: NDArray[np.float64]
    ) -> str:
        keys, groups = numpy_group_by(group=fitness, by=operators)
        mean_fit = np.array(list(map(np.mean, groups)))
        fittest_operator = keys[np.argmax(mean_fit)]
        return fittest_operator

    def _update_data(self: SelfCGA) -> None:
        self._update_fittest(self._population_g_i, self._population_ph_i, self._fitness_i)
        self._update_stats(
            population_g=self._population_ph_i,
            fitness_max=self._thefittest._fitness,
            s_proba=self._selection_proba,
            c_proba=self._crossover_proba,
            m_proba=self._mutation_proba,
        )

    def _adapt(self: SelfCGA) -> None:
        s_fittest_oper = self._find_fittest_operator(
            self._selection_operators, self._fitness_scale_i
        )
        self._selection_proba = self._get_new_proba(self._selection_proba, s_fittest_oper)

        c_fittest_oper = self._find_fittest_operator(
            self._crossover_operators, self._fitness_scale_i
        )
        self._crossover_proba = self._get_new_proba(self._crossover_proba, c_fittest_oper)

        m_fittest_oper = self._find_fittest_operator(
            self._mutation_operators, self._fitness_scale_i
        )
        self._mutation_proba = self._get_new_proba(self._mutation_proba, m_fittest_oper)

    def _get_new_population(self: SelfCGA) -> None:
        self._selection_operators = self._choice_operators(proba_dict=self._selection_proba)
        self._crossover_operators = self._choice_operators(proba_dict=self._crossover_proba)
        self._mutation_operators = self._choice_operators(proba_dict=self._mutation_proba)

        self._population_g_i = np.array(
            [
                self._get_new_individ_g(
                    self._selection_operators[i],
                    self._crossover_operators[i],
                    self._mutation_operators[i],
                )
                for i in range(self._pop_size)
            ],
            dtype=self._population_g_i.dtype,
        )



from __future__ import annotations

import random
from collections import defaultdict
from inspect import signature
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import NDArray

from ..tools import find_end_subtree_from_i
from ..tools import find_id_args_from_i
from ..tools import get_levels_tree_from_i


FUNCTIONAL_COLOR_CODE = (1, 0.72, 0.43, 1)
TERMINAL_COLOR_CODE = (0.21, 0.76, 0.56, 1)


class Operator:
    def __init__(self, formula: str, name: str, sign: str) -> None:
        self._formula = formula
        self.__name__ = name
        self._sign = sign

    def _write(self, *args: Any) -> str:
        formula = self._formula.format(*args)
        return formula

    def __call__(self, *args: Any) -> None:
        pass


class Node:
    def __init__(self, value: Any, name: str, sign: str, n_args: int) -> None:
        self._value = value
        self._name = name
        self._sign = sign
        self._n_args = n_args

    def __str__(self) -> str:
        return str(self._sign)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Node):
            return NotImplemented
        return self._name == other._name


class FunctionalNode(Node):
    def __init__(self, value: Any, sign: Optional[str] = None) -> None:
        Node.__init__(
            self,
            value=value,
            name=value.__name__,
            sign=sign or value._sign,
            n_args=len(signature(value.__call__).parameters),
        )


class TerminalNode(Node):
    def __init__(self, value: Any, name: str) -> None:
        Node.__init__(self, value=value, name=name, sign=name, n_args=0)


class EphemeralConstantNode(TerminalNode):
    def __init__(self, value: Any, name: str) -> None:
        TerminalNode.__init__(self, value=value, name=name)


class EphemeralNode(Node):
    def __init__(self, generator: Callable) -> None:
        self._generator = generator
        Node.__init__(
            self,
            value=self._generator,
            name=str(self._generator.__name__),
            sign=str(self._generator.__name__),
            n_args=0,
        )

    def __call__(self) -> EphemeralConstantNode:
        value = self._generator()
        node = EphemeralConstantNode(value=value, name=str(value))
        return node


class UniversalSet:
    def __init__(
        self,
        functional_set: Tuple[FunctionalNode, ...],
        terminal_set: Tuple[Union[TerminalNode, EphemeralNode], ...],
    ) -> None:
        self._functional_set = self._define_functional_set(functional_set)
        self._terminal_set = tuple(terminal_set)

    def _define_functional_set(self, functional_set: Tuple[FunctionalNode, ...]) -> defaultdict:
        _functional_set_list = defaultdict(list, {-1: list(functional_set)})
        _functional_set_tuple = defaultdict(tuple)
        for unit in functional_set:
            _functional_set_list[unit._n_args].append(unit)
        for key, value in _functional_set_list.items():
            _functional_set_tuple[key] = tuple(value)
        return _functional_set_tuple

    def _random_terminal_or_ephemeral(self) -> Union[EphemeralConstantNode, TerminalNode]:
        choosen = random.choice(self._terminal_set)
        if isinstance(choosen, EphemeralNode):
            return choosen()
        else:
            return choosen

    def _random_functional(self, n_args: int = -1) -> FunctionalNode:
        n_args_functionals = self._functional_set[n_args]
        node = random.choice(n_args_functionals)
        return node


class Tree:
    def __init__(
        self,
        nodes: List[Union[FunctionalNode, TerminalNode, EphemeralNode]],
        n_args: Optional[Union[List[int], NDArray[np.int64]]] = None,
    ) -> None:
        self._nodes = nodes
        if n_args is None:
            self._n_args = self._init_n_args()
        else:
            self._n_args = np.array(n_args, dtype=np.int64)

    def _init_n_args(self) -> NDArray[np.int64]:
        n_args = np.empty(shape=len(self._nodes), dtype=np.int64)
        for i, node_i in enumerate(self._nodes):
            n_args[i] = node_i._n_args
        return n_args

    def __len__(self) -> int:
        return len(self._nodes)

    def __str__(self) -> str:
        pack: List[str] = []
        for node in reversed(self._nodes):
            args = []
            for _ in range(node._n_args):
                args.append(pack.pop())
            if isinstance(node, FunctionalNode):
                pack.append(node._value._write(*args))
            else:
                pack.append(node._name)
        return pack[0]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Tree):
            return NotImplemented
        else:
            if len(self) != len(other):
                return False
            else:
                for node_1, node_2 in zip(self._nodes, other._nodes):
                    if node_1 != node_2:
                        return False
            return True

    def __call__(self) -> Any:
        pack: Any = []
        for node in reversed(self._nodes):
            args = []
            for _ in range(node._n_args):
                args.append(pack.pop())
            if isinstance(node, FunctionalNode):
                pack.append(node._value(*args))
            else:
                pack.append(node._value)
        return pack[0]

    def copy(self) -> Tree:
        return Tree(self._nodes.copy(), self._n_args.copy())

    def set_terminals(self, **kwargs: Any) -> Tree:
        tree_copy = self.copy()
        for i, node in enumerate(tree_copy._nodes):
            if isinstance(node, TerminalNode):
                for name, value in kwargs.items():
                    if node._name == name:
                        tree_copy._nodes[i] = TerminalNode(value=value, name=node._name)
        return tree_copy

    def subtree_id(self, index: int) -> Tuple[int, int]:
        n_index = find_end_subtree_from_i(np.int64(index), self._n_args)
        return (index, n_index)

    def subtree(self, index: int) -> Tree:
        n_index = find_end_subtree_from_i(np.int64(index), self._n_args)
        new_tree = Tree(self._nodes[index:n_index].copy(), self._n_args[index:n_index].copy())
        return new_tree

    def concat(self, index: int, other_tree: Tree) -> Tree:
        to_return = self.copy()
        left, right = self.subtree_id(index)
        to_return._nodes[left:right] = other_tree._nodes.copy()
        to_return._n_args = np.r_[
            to_return._n_args[:left], other_tree._n_args.copy(), to_return._n_args[right:]
        ]
        return to_return

    def get_args_id(self, index: int) -> NDArray[np.int64]:
        args_id = find_id_args_from_i(np.int64(index), self._n_args)
        return args_id

    def get_levels(self, index: int) -> NDArray[np.int64]:
        return get_levels_tree_from_i(np.int64(index), self._n_args)

    def get_max_level(self) -> np.int64:
        return max(self.get_levels(0))

    def get_graph(self, keep_id: bool = False) -> Dict:
        pack: List[int] = []
        edges = []
        nodes = []
        labels = {}
        for i, node in enumerate(reversed(self._nodes)):
            index = len(self) - i - 1
            if keep_id:
                labels[index] = str(index) + ". " + node._sign[:60]
            else:
                labels[index] = node._sign[:60]

            nodes.append(index)

            for _ in range(node._n_args):
                edges.append((index, len(self) - pack.pop() - 1))
            pack.append(i)

        edges.reverse()
        nodes.reverse()

        levels = self.get_levels(0)
        colors = np.zeros(shape=(len(nodes), 4))
        pos = np.zeros(shape=(len(self), 2))
        for i, lvl_i in enumerate(levels):
            total = 0
            cond = lvl_i == np.array(levels)
            h = 1 / (1 + np.sum(cond))
            arange = np.arange(len(pos))[cond]

            for a_j in arange:
                total += h
                pos[a_j][0] = total

            pos[i][1] = -lvl_i

            if isinstance(self._nodes[i], FunctionalNode):
                colors[i] = FUNCTIONAL_COLOR_CODE
            else:
                colors[i] = TERMINAL_COLOR_CODE

        to_return = {"edges": edges, "labels": labels, "nodes": nodes, "pos": pos, "colors": colors}
        return to_return
